# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Geoips plugin for driving pyresample Gaussian interpolation."""

import logging

import xarray

from geoips.plugins.modules.interpolators.utils.interp_pyresample import (
    interp_kd_tree,
    get_data_box_definition,
)
from geoips.geoips_utils import copy_standard_metadata, remove_unsupported_kwargs

LOG = logging.getLogger(__name__)

interface = "interpolators"
family = "2d"
name = "interp_gauss"


def call(
    area_def,
    input_xarray,
    output_xarray,
    varlist,
    array_num=None,
    sigmaval=None,
    drop_nan=False,
    **kwargs,
):
    """Pyresample interp_kd_tree gaussian interpolation GeoIPS plugin."""
    LOG.info(
        "Interpolating using standard scifile register method: "
        "kd_tree gauss sigmaval %s",
        sigmaval,
    )

    goodinds = None

    vars_to_interp = []
    if array_num is not None:
        lons = input_xarray["longitude"][:, :, array_num]
        lats = input_xarray["latitude"][:, :, array_num]
        if drop_nan is True:
            # Make sure we get any bad lat/lon values
            goodinds = ~input_xarray["latitude"][:, :, array_num].isnull()
            goodinds = ~input_xarray["longitude"][:, :, array_num].isnull() & goodinds
        for varname in varlist:
            if drop_nan is True:
                # Cumulative mask over all channels. What could go wrong?
                goodinds = ~input_xarray[varname][:, :, array_num].isnull() & goodinds
                vars_to_interp += [input_xarray[varname][:, :, array_num]]
            else:
                vars_to_interp += [
                    input_xarray[varname].to_masked_array()[:, :, array_num]
                ]
    else:
        lons = input_xarray["longitude"]
        lats = input_xarray["latitude"]
        if drop_nan is True:
            # Make sure we get any bad lat/lon values
            goodinds = ~input_xarray["latitude"].isnull()
            goodinds = ~input_xarray["longitude"].isnull() & goodinds
        for varname in varlist:
            if drop_nan is True:
                # Cumulative mask over all channels. What could go wrong?
                goodinds = ~input_xarray[varname].isnull() & goodinds
                vars_to_interp += [input_xarray[varname]]
            else:
                vars_to_interp += [input_xarray[varname].to_masked_array()]

    if drop_nan is True:
        final_vars_to_interp = []
        lats = lats.where(goodinds, drop=True)
        lons = lons.where(goodinds, drop=True)
        for var in vars_to_interp:
            final_vars_to_interp += [var.where(goodinds, drop=True).to_masked_array()]
        vars_to_interp = final_vars_to_interp
    lats = lats.to_masked_array()
    lons = lons.to_masked_array()
    # Use standard scifile / pyresample registration
    data_box_definition = get_data_box_definition(input_xarray.source_name, lons, lats)

    # Set s default value of igmaval as 10000

    if sigmaval is None:
        sigmaval = 10000

    kd_kwargs = remove_unsupported_kwargs(interp_kd_tree, kwargs)
    interp_data = interp_kd_tree(
        vars_to_interp,
        area_def,
        data_box_definition,
        input_xarray.interpolation_radius_of_influence,
        interp_type="gauss",
        sigmas=sigmaval,
        **kd_kwargs,
    )
    if output_xarray is None:
        output_xarray = xarray.Dataset()
    if "latitude" not in output_xarray.variables.keys():
        interp_lons, interp_lats = area_def.get_lonlats()
        output_xarray["latitude"] = xarray.DataArray(interp_lats)
        output_xarray["longitude"] = xarray.DataArray(interp_lons)
    copy_standard_metadata(input_xarray, output_xarray, force=False)
    output_xarray.attrs["registered_dataset"] = True
    output_xarray.attrs["area_definition"] = area_def

    for ind in range(len(varlist)):
        output_xarray[varlist[ind]] = xarray.DataArray(interp_data[ind])

    return output_xarray



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS CLI "run" command.

Runs the appropriate script based on the args provided.
"""

from geoips.commandline.args import add_args
from geoips.commandline.run_procflow import main
from geoips.commandline.geoips_command import GeoipsCommand, GeoipsExecutableCommand
from geoips.utils.context_managers import import_optional_dependencies

data_fusion_installed = False

with import_optional_dependencies(loglevel="info"):
    """Attempt to import data_fusion_args from Data Fusion."""
    # NOTE: loglevel is set to 'info' here so we don't get the constant output of
    # 'Failed to import data_fusion.commandline at /path/to/geoips/geoips/commandline/geoips_run.py:19. If you need it, install it.' # NOQA
    # every time we use the CLI. If a user needs the CLI, I'm assuming they'll know to
    # install it.
    from data_fusion.commandline.args import add_args as data_fusion_add_args

    data_fusion_installed = True


class GeoipsRunConfigBased(GeoipsExecutableCommand):
    """Run Command for executing the config based process-workflow (procflow)."""

    name = "config_based"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the run-subparser for the 'run config_based' Command."""
        add_args(parser=self.parser, legacy=self.legacy)

    def __call__(self, args):
        """Run the provided GeoIPS command.

        Parameters
        ----------
        args: Namespace()
            - The argument namespace to parse through
        """
        if args.procflow is None and self.legacy:
            err_str = (
                "Deprecated, Legacy 'run_procflow' call was used and --procflow "
                "flag wasn't specified. Please either specify which procflow is "
                "being executed via '--procflow <procflow_name>' or use the "
                "supported procflow call 'geoips run <procflow_name>'"
            )
            self.parser.error(err_str)
        elif args.procflow is None:
            # If None, set to 'config_based'. We don't want users to have to specify
            # what procflow will be used as it is specified in
            # 'geoips run config_based'.
            args.procflow = "config_based"
        main(ARGS=args)


class GeoipsRunDataFusion(GeoipsExecutableCommand):
    """Run Command for executing the data fusion process-workflow (procflow)."""

    name = "data_fusion"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the run-subparser for the 'run data_fusion' Command."""
        if data_fusion_installed:
            data_fusion_add_args(parser=self.parser, legacy=self.legacy)

    def __call__(self, args):
        """Run the provided GeoIPS command.

        Parameters
        ----------
        args: Namespace()
            - The argument namespace to parse through
        """
        if data_fusion_installed:
            if args.procflow is None and self.legacy:
                err_str = (
                    "Deprecated, Legacy 'data_fusion_procflow' call was used and "
                    "--procflow flag wasn't specified. Please either specify which "
                    "procflow is being executed via '--procflow <procflow_name>' or "
                    "use the supported procflow call 'geoips run <procflow_name>'"
                )
                self.parser.error(err_str)
            elif args.procflow is None:
                # If None, set to 'data_fusion'. We don't want users to have to specify
                # what procflow will be used as it is specified in
                # 'geoips run data_fusion'.
                args.procflow = "data_fusion"
            main(ARGS=args)
        else:
            print(
                "Data fusion is not installed. If you want to run this type of "
                "functionality, please clone and install "
                "https://github.com/NRLMMD-GEOIPS/data_fusion.git"
            )


class GeoipsRunSingleSource(GeoipsExecutableCommand):
    """Run Command for executing the single source process-workflow (procflow)."""

    name = "single_source"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the run-subparser for the 'run single_source' Command."""
        add_args(parser=self.parser, legacy=self.legacy)

    def __call__(self, args):
        """Run the provided GeoIPS command.

        Parameters
        ----------
        args: Namespace()
            - The argument namespace to parse through
        """
        if args.procflow is None and self.legacy:
            err_str = (
                "Deprecated, Legacy 'run_procflow' call was used and --procflow "
                "flag wasn't specified. Please either specify which procflow is "
                "being executed via '--procflow <procflow_name>' or use the "
                "supported procflow call 'geoips run <procflow_name>'"
            )
            self.parser.error(err_str)
        elif args.procflow is None:
            # If None, set to 'single_source'. We don't want users to have to specify
            # what procflow will be used as it is specified in
            # 'geoips run single_source'.
            args.procflow = "single_source"
        main(ARGS=args)


class GeoipsRun(GeoipsCommand):
    """Run Command for running process-workflows (procflows)."""

    name = "run"
    command_classes = [
        GeoipsRunSingleSource,
        GeoipsRunDataFusion,
        GeoipsRunConfigBased,
    ]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Full disk image matplotlib-based output format."""

import logging

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "image_overlay"
name = "full_disk_image"


def call(
    area_def,
    xarray_obj,
    product_name,
    output_fnames,
    clean_fname=None,
    product_name_title=None,
    mpl_colors_info=None,
    feature_annotator=None,
    gridline_annotator=None,
    product_datatype_title=None,
    bg_data=None,
    bg_mpl_colors_info=None,
    bg_xarray=None,
    bg_product_name_title=None,
    bg_datatype_title=None,
    remove_duplicate_minrange=None,
):
    """Plot full disk image."""
    if product_name_title is None:
        product_name_title = product_name

    success_outputs = []
    plot_data = xarray_obj[product_name].to_masked_array()
    from geoips.image_utils.mpl_utils import create_figure_and_main_ax_and_mapobj
    from geoips.image_utils.mpl_utils import (
        save_image,
        plot_overlays,
        create_colorbar,
    )
    from geoips.image_utils.mpl_utils import get_title_string_from_objects, set_title

    frame_clr = None
    # If a gridline_annotator plugin was supplied, attempt to get the frame background
    # color. Otherwise, just keep it as None.
    if gridline_annotator:
        frame_clr = gridline_annotator.get("spec", {}).get("background")

    if hasattr(area_def, "x_size"):
        x_size = area_def.x_size
        y_size = area_def.y_size
    else:
        x_size = area_def.get_lonlats()[0].shape[0]
        y_size = area_def.get_lonlats()[0].shape[1]

    import cartopy

    if hasattr(area_def, "proj_dict"):
        # mapobj = cartopy.crs.Orthographic(area_def.proj_dict['lon_0'],
        #                                   area_def.proj_dict['lat_0'])
        mapobj = cartopy.crs.Geostationary(area_def.proj_dict["lon_0"])
    else:
        # mapobj = cartopy.crs.Orthographic(area_def.sector_info['clon'],
        #                                   area_def.sector_info['clat'])
        mapobj = cartopy.crs.Geostationary(area_def.sector_info["clon"])
    fig, main_ax, mapobj = create_figure_and_main_ax_and_mapobj(
        x_size,
        y_size,
        area_def,
        existing_mapobj=mapobj,
        frame=frame_clr,
    )

    # Plot the actual data on a map

    main_ax.imshow(
        plot_data,
        transform=area_def.to_cartopy_crs(),
        # extent=area_def.area_extent_ll,
        cmap=mpl_colors_info["cmap"],
        norm=mpl_colors_info["norm"],
    )

    # Set the title for final image
    title_string = get_title_string_from_objects(
        area_def,
        xarray_obj,
        product_name_title,
        product_datatype_title=product_datatype_title,
        bg_xarray=bg_xarray,
        bg_product_name_title=bg_product_name_title,
        bg_datatype_title=bg_datatype_title,
    )
    set_title(main_ax, title_string, y_size)

    if mpl_colors_info["colorbar"] is True:
        # Create the colorbar to match the mpl_colors
        create_colorbar(fig, mpl_colors_info)

    # Plot gridlines and feature overlays
    plot_overlays(
        mapobj,
        main_ax,
        area_def,
        feature_annotator=feature_annotator,
        gridline_annotator=gridline_annotator,
    )

    if output_fnames is not None:
        for annotated_fname in output_fnames:
            # Save the final image
            success_outputs += save_image(
                fig,
                annotated_fname,
                is_final=True,
                image_datetime=xarray_obj.start_datetime,
                remove_duplicate_minrange=remove_duplicate_minrange,
            )

    return success_outputs



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Sector interface module."""

from geoips.interfaces.base import BaseYamlPlugin, BaseYamlInterface

# Uncomment when ready to switch from JsonSchema to Pydantic
# from geoips.pydantic.sectors import SectorPluginModel

# Commenting these out for PR #260
# Will work on this again after the 2023 workshop
#
# def center_to_area_definition(sector):
#     """Return a pyresample AreaDefinition for the input sector.
#
#     The input sector must supply location information in the "center" format.
#     """
#     if not sector.family.startswith("center"):
#         raise ValueError("Sector does not supply location as center coordinates.")
#     raise NotImplementedError
#
#
# def corners_to_area_definition(sector):
#     """Return a pyresample AreaDefinition for the input sector.
#
#     The input sector must supply location information in the "corners" format.
#     """
#     if not sector.family.startswith("center"):
#         raise ValueError("Sector does not supply location as corner coordinates.")
#
#     ad_info = {
#         "area_id": sector.name,
#         "projection": {
#             "units": "m",
#             "a": 6371228.0,
#             "proj": sector.spec.projection,
#             "lat_0": sector.spec.center.lat,
#             "lon_0": sector.spec.center.lon,
#         },
#         "width": sector.shape[0],
#         "height": sector.shape[1],
#         "resolution": [sector.resolution, sector.resolution],
#         "center": [0, 0],
#     }
#
#     ad = create_area_def(**ad_info)
#     raise NotImplementedError


class SectorPluginBase(BaseYamlPlugin):
    """The base class for all sector plugins.

    This class provides the functionality specific to sector plugin classes. It should
    not be instantiated directly. To access sector plugins, use
    `geoips.interfaces.sectors`.
    """

    @property
    def area_definition(self):
        """Return the pyresample AreaDefinition for the sector."""
        # if self.family.startswith(("area_definition", "generated")):
        from pyresample import load_area
        from importlib.resources import files

        if self.family.startswith("area_definition"):
            abspath = str(files(self.package) / self.relpath)
            ad = load_area(abspath, "spec")
        # elif self.family.startswith("center"):
        #     ad = center_to_area_definition(self)
        # elif self.family.startswith("corners"):
        #     ad = corners_to_area_definition(self)
        return ad

    def create_test_plot(self, fname, return_fig_ax_map=False):
        """Create a test PNG image for this sector."""
        from geoips.image_utils.mpl_utils import create_figure_and_main_ax_and_mapobj
        from cartopy import feature as cfeature

        fig, ax, mapobj = create_figure_and_main_ax_and_mapobj(
            self.area_definition.shape[1],
            self.area_definition.shape[0],
            self.area_definition,
            noborder=True,
        )
        ax.add_feature(cfeature.COASTLINE)
        ax.add_feature(cfeature.BORDERS)
        if fname is not None:
            fig.savefig(fname)
        if return_fig_ax_map:
            return fig, ax, mapobj


class SectorsInterface(BaseYamlInterface):
    """Interface for sector plugins."""

    name = "sectors"
    plugin_class = SectorPluginBase
    # Uncomment when ready to switch from JsonSchema to Pydantic
    # validator = SectorPluginModel


sectors = SectorsInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Matplotlib-based windbarb clean image output (no overlays or backgrounds)."""

import logging

from geoips.plugins.modules.output_formatters.imagery_windbarbs import (
    output_clean_windbarbs,
    format_windbarb_data,
)

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "image"
name = "imagery_windbarbs_clean"


def call(
    area_def,
    xarray_obj,
    product_name,
    output_fnames,
    product_name_title=None,
    mpl_colors_info=None,
    existing_image=None,
    remove_duplicate_minrange=None,
    fig=None,
    main_ax=None,
    mapobj=None,
):
    """Plot clean windbarb imagery on matplotlib figure."""
    formatted_data_dict = format_windbarb_data(xarray_obj, product_name)

    success_outputs = output_clean_windbarbs(
        area_def,
        output_fnames,
        mpl_colors_info,
        xarray_obj.start_datetime,
        formatted_data_dict,
        fig,
        main_ax,
        mapobj,
    )

    return success_outputs



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Interpolation interface will be deprecated v2.0.

Wrapper functions for geoips output_config specifications.

This functionality will be replaced with a class-based implementation v2.0,
and deprecated at that time.
"""
import logging
from geoips.interfaces import products
from geoips.interfaces import colormappers
from geoips.interfaces import feature_annotators
from geoips.interfaces import gridline_annotators

LOG = logging.getLogger(__name__)


# ### Output config dictionaries ###
def is_valid_output_config(output_config_dict):
    """Interface will be deprecated v2.0.

    Check that requested output_config dictionary is properly formatted.

    The dictionary of output_config
    parameters fully determines the outputs required for a given set of
    data files.

    Dictionary of output_config parameters currently specified by a
    full path to a YAML file:
    and requested via commandline with:
    ``--output_config <full_path_to_YAML_output_config>``

    Parameters
    ----------
    output_config_dict : dict
        Dictionary of output config parameters

    Returns
    -------
    is_valid : bool
        * True if ``output_config_dict`` is a properly formatted dictionary of
          output parameters.
        * False if output_config_dict:
            * does not contain supported ``output_config_type``,
            * does not contain all ``required`` fields,
            * contains non-supported ``optional`` fields

    Notes
    -----
    output_config_types currently one of:
        * ``single_source``
        * ``fused``
    """
    required_keys = {}
    optional_keys = {}

    required_keys["single_source"] = {
        "output_config_type": [],
        "reader_name": [],
        "file_names": [],
        "sectored_read": [],
        "available_sectors": [],
        "outputs": [
            "requested_sector_type",
            "output_formatter",
            "filename_formatters",
            "product_names",
        ],
    }

    optional_keys["single_source"] = {
        "available_sectors": [
            "sector_list",
            "trackfile_parser",
            "trackfiles",
            "tc_spec_template",
        ],
        "outputs": [
            "remove_duplicates",
            "metadata_filename_formatter",
            "metadata_filename_formatters",
            "metadata_filename_formatters_kwargs",
            "minimum_coverages",
            "minimum_coverage",
            "compare_path",
            "feature_annotator",
            "gridline_annotator",
        ],
    }

    required_keys["fused"] = required_keys["single_source"].copy()
    optional_keys["fused"] = optional_keys["single_source"].copy()
    optional_keys["fused"] = {
        "fuse_files": [],
        "fuse_reader": [],
        "fuse_product": [],
        "outputs": {"background_products": ["config_names", "product_names"]},
    }

    if "output_config_type" not in output_config_dict:
        LOG.error(
            "INVALID OUTPUT CONFIG '%s': "
            "'output_config_type' must be defined within output config dictionary",
            output_config_dict.keys(),
        )
        return False

    if output_config_dict["output_config_type"] not in required_keys:
        LOG.error(
            "INVALID OUTPUT CONFIG '%s': "
            "'output_config_type' in output config dictionary must be one of '%s'",
            output_config_dict.keys(),
            list(required_keys.keys()),
        )
        return False

    output_config_type = output_config_dict["output_config_type"]

    # If we don't have all of the required keys, return False
    if not set(required_keys[output_config_type]).issubset(set(output_config_dict)):
        LOG.error(
            "INVALID OUTPUT CONFIG '%s': "
            "'%s' output config dictionary must contain the following fields: '%s'",
            output_config_dict.keys(),
            output_config_type,
            list(required_keys[output_config_type]),
        )
        return False

    # If we have non-allowed keys, return False
    if not set(output_config_dict).issubset(
        required_keys[output_config_type] + optional_keys[output_config_type]
    ):
        LOG.error(
            "INVALID OUTPUT CONFIG'%s': "
            "Unknown fields in '%s' output config dictionary: '%s'",
            output_config_dict.keys(),
            output_config_type,
            set(output_config_dict).difference(
                required_keys[output_config_type] + optional_keys[output_config_type]
            ),
        )
        return False

    # Probably need to add in the nested fields too...

    # If we get here, then the output config dictionary format is valid.
    return True


def get_output_config_type(output_config_dict):
    """Interface will be deprecated v2.0.

    Retrieve output_config_type of the passed output_config_dict, found in:
           output_config_dict['output_config_type']

    See: geoips.dev.output_config.is_valid_output_config for full list of
    supported output_config types.

    Parameters
    ----------
    output_config_dict : dict
        dictionary of complete output config parameters

    Returns
    -------
        (str) : output_config type, found in output_config_dict['output_config_type']
    """
    return output_config_dict["output_config_type"]


def get_filename_formatters(output_dict):
    """Interface will be deprecated v2.0."""
    if "filename_formatter" in output_dict and isinstance(
        output_dict["filename_formatter"], str
    ):
        return [output_dict["filename_formatter"]]
    else:
        return output_dict["filename_formatters"]


def get_output_formatter(output_dict):
    """Interface will be deprecated v2.0."""
    return output_dict["output_formatter"]


def get_metadata_output_formatter(output_dict):
    """Interface will be deprecated v2.0."""
    if "metadata_output_formatter" in output_dict:
        return output_dict["metadata_output_formatter"]
    return None


def get_metadata_filename_formatter(filename_formatter, output_dict):
    """Interface will be deprecated v2.0."""
    fname_format = None

    if "metadata_filename_formatter" in output_dict and isinstance(
        output_dict["metadata_filename_formatter"], str
    ):
        fname_format = output_dict["metadata_filename_formatter"]

    elif (
        "metadata_filename_formatters" in output_dict
        and filename_formatter in output_dict["metadata_filename_formatters"]
    ):
        fname_format = output_dict["metadata_filename_formatters"][filename_formatter]

    elif (
        "metadata_filename_formatters" in output_dict
        and "all" in output_dict["metadata_filename_formatters"]
    ):
        fname_format = output_dict["metadata_filename_formatters"]["all"]

    return fname_format


def get_minimum_coverage(product_name, output_dict):
    """Interface will be deprecated v2.0."""
    minimum_coverage = None

    if (
        "minimum_coverage" in output_dict
        and output_dict["minimum_coverage"] is not None
    ):
        minimum_coverage = float(output_dict["minimum_coverage"])

    elif (
        "minimum_coverages" in output_dict
        and product_name in output_dict["minimum_coverages"]
    ):
        minimum_coverage = float(output_dict["minimum_coverages"][product_name])

    elif (
        "minimum_coverages" in output_dict and "all" in output_dict["minimum_coverages"]
    ):
        minimum_coverage = float(output_dict["minimum_coverages"]["all"])

    if minimum_coverage is not None and (
        minimum_coverage < 0 or minimum_coverage > 100
    ):
        raise ValueError(
            "Minimum coverage must be percentage between 0 and 100, "
            f"{minimum_coverage} invalid"
        )

    return minimum_coverage


def get_filename_formatter_kwargs(filename_formatter, output_dict):
    """Interface will be deprecated v2.0.

    Return dictionary of filename_formatters_kwargs.

    based on what was passed in via the YAML output config
    dictionary, as well as default kwargs.

    If "filename_formatter_kwargs (singular) is passed command line, use
    that to override ALL filename_formatters_kwargs specified in YAML output config.
    """
    filename_formatter_kwargs = {}

    # YAML output config arguments
    if (
        "filename_formatters_kwargs" in output_dict
        and "all" in output_dict["filename_formatters_kwargs"]
    ):
        filename_formatter_kwargs = output_dict["filename_formatters_kwargs"]["all"]
    if (
        "filename_formatters_kwargs" in output_dict
        and filename_formatter in output_dict["filename_formatters_kwargs"]
    ):
        filename_formatter_kwargs = output_dict["filename_formatters_kwargs"][
            filename_formatter
        ]
    # Command line argument overrides all
    if "filename_formatter_kwargs" in output_dict:
        filename_formatter_kwargs = output_dict["filename_formatter_kwargs"]

    filename_formatter_kwargs["output_dict"] = output_dict

    return filename_formatter_kwargs


def get_metadata_filename_formatter_kwargs(filename_formatter, output_dict):
    """Interface will be deprecated v2.0.

    Return dictionary of filename_formatters_kwargs.

    based on what was passed in via the YAML output config
    dictionary, as well as default kwargs
    """
    metadata_filename_formatter_kwargs = {}

    if (
        "metadata_filename_formatters_kwargs" in output_dict
        and "all" in output_dict["metadata_filename_formatters_kwargs"]
    ):
        metadata_filename_formatter_kwargs = output_dict[
            "metadata_filename_formatters_kwargs"
        ]["all"]

    if (
        "metadata_filename_formatters_kwargs" in output_dict
        and filename_formatter in output_dict["metadata_filename_formatters_kwargs"]
    ):
        metadata_filename_formatter_kwargs = output_dict[
            "metadata_filename_formatters_kwargs"
        ][filename_formatter]

    metadata_filename_formatter_kwargs["output_dict"] = output_dict

    return metadata_filename_formatter_kwargs


def set_lonlat_spacing(gridline_annotator, area_def):
    """Interface will be deprecated v2.0."""
    if (
        gridline_annotator is None
        or "grid_lat_spacing" not in gridline_annotator.keys()
        or "grid_lon_spacing" not in gridline_annotator.keys()
        or gridline_annotator["grid_lon_spacing"] == "auto"
        or gridline_annotator["grid_lat_spacing"] == "auto"
    ):
        from pyresample import utils

        minlat = area_def.area_extent_ll[1]
        maxlat = area_def.area_extent_ll[3]
        minlon = utils.wrap_longitudes(area_def.area_extent_ll[0])
        maxlon = utils.wrap_longitudes(area_def.area_extent_ll[2])
        if minlon > maxlon and maxlon < 0:
            maxlon = maxlon + 360
        lon_extent = maxlon - minlon
        lat_extent = maxlat - minlat

        if lon_extent > 5:
            lon_spacing = int(lon_extent / 5)
        elif lon_extent > 2.5:
            lon_spacing = 1
        else:
            lon_spacing = lon_extent / 5.0

        if lat_extent > 5:
            lat_spacing = int(lat_extent / 5)
        elif lat_extent > 2.5:
            lat_spacing = 1
        else:
            lat_spacing = lat_extent / 5.0

        LOG.info("lon_extent: %s, lon_spacing: %s", lon_extent, lon_spacing)
        gridline_annotator["grid_lat_spacing"] = lat_spacing
        gridline_annotator["grid_lon_spacing"] = lon_spacing

    return gridline_annotator


def get_output_formatter_kwargs(
    output_dict,
    xarray_obj=None,
    area_def=None,
    sector_type=None,
    bg_files=None,
    bg_xarrays=None,
    bg_product_name=None,
):
    """Interface will be deprecated v2.0."""
    output_formatter_kwargs = {}
    if "output_formatter_kwargs" in output_dict:
        output_formatter_kwargs = output_dict["output_formatter_kwargs"].copy()

    if (
        "gridline_annotator" in output_dict
        and output_dict["gridline_annotator"] is not None
    ):
        gridline_annotator = gridline_annotators.get_plugin(
            output_dict["gridline_annotator"]
        )
        gridline_annotator = set_lonlat_spacing(gridline_annotator, area_def)
        output_formatter_kwargs["gridline_annotator"] = gridline_annotator

    if (
        "feature_annotator" in output_dict
        and output_dict["feature_annotator"] is not None
    ):
        feature_annotator = feature_annotators.get_plugin(
            output_dict["feature_annotator"]
        )
        output_formatter_kwargs["feature_annotator"] = feature_annotator

    if bg_files and "background_products" in output_dict and sector_type in bg_xarrays:
        output_formatter_kwargs["bg_xarray"] = bg_xarrays[sector_type]
        output_formatter_kwargs["bg_data"] = output_formatter_kwargs["bg_xarray"][
            bg_product_name
        ].to_masked_array()
        output_formatter_kwargs["bg_product_name_title"] = bg_product_name
        output_formatter_kwargs["bg_mpl_colors_info"] = None
        if bg_product_name:
            bg_prod_plugin = products.get_plugin(
                output_formatter_kwargs["bg_xarray"].source_name,
                bg_product_name,
                output_dict.get("product_spec_override"),
            )
            bg_cmap_plugin = colormappers.get_plugin(
                bg_prod_plugin["spec"]["colormapper"]["plugin"]["name"]
            )
            bg_cmap_args = bg_prod_plugin["spec"]["colormapper"]["plugin"]["arguments"]

            output_formatter_kwargs["bg_mpl_colors_info"] = bg_cmap_plugin(
                **bg_cmap_args
            )

    output_formatter_kwargs["output_dict"] = output_dict

    return output_formatter_kwargs


def get_metadata_output_formatter_kwargs(output_dict):
    """Interface will be deprecated v2.0."""
    metadata_output_formatter_kwargs = {}
    if "metadata_output_formatter_kwargs" in output_dict:
        metadata_output_formatter_kwargs = output_dict[
            "metadata_output_formatter_kwargs"
        ].copy()
        metadata_output_formatter_kwargs["output_dict"] = output_dict
    return metadata_output_formatter_kwargs


def produce_current_time(config_dict, metadata_xobj, output_dict_keys=None):
    """Interface will be deprecated v2.0.

    Determine if the current data file needs to be processed,
        based on the requested times.
    If output_dict_key is included, apply to only the currently
        requested output_dict.
    If output_dict_key is None, check ALL outputs to determine if
        ANY need the current time.
    """
    if output_dict_keys is None:
        output_dict_keys = config_dict["outputs"].keys()

    retval = False
    for output_dict_key in output_dict_keys:
        LOG.info("CHECKING for required time in output_dict: %s", output_dict_key)
        output_dict = config_dict["outputs"][output_dict_key]
        # If 'produce_times' is not in the current output_dict,
        # it means we DO NOT filter based on time, so we
        # will need this file time processed.
        if "produce_times" not in output_dict:
            LOG.info(
                'WILL PROCESS "produce_times" not included in output_dict: %s',
                output_dict_key,
            )
            retval = True
        elif "required_minutes" in output_dict["produce_times"]:
            if (
                metadata_xobj.start_datetime.minute
                in output_dict["produce_times"]["required_minutes"]
            ):
                LOG.info(
                    "WILL PROCESS produce_times['required_minutes']=%s "
                    "includes current minute '%s', '%s'",
                    output_dict["produce_times"]["required_minutes"],
                    metadata_xobj.start_datetime.minute,
                    metadata_xobj.start_datetime,
                )
                retval = True

    return retval


def test_output_config_interface(output_config_dict):
    """Interface will be deprecated v2.0.

    Finds and opens every product params dict available within the current
    geoips instantiation

    See geoips.dev.output_config.is_valid_output_config?
        for a list of available product params dict types and
        associated call signatures / return values.

    Returns
    -------
    list
        List of all successful output_config information
    """
    out_dict = {}
    # LOG.info('Checking source %s', source_name)
    out_dict["validity_check"] = is_valid_output_config(output_config_dict)
    out_dict["get_output_config_type"] = get_output_config_type(output_config_dict)

    # Need to loop through all the outputs and available sectors...

    return out_dict



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Coverage check routine for windbarb xarrays."""

import logging

LOG = logging.getLogger(__name__)

interface = "coverage_checkers"
family = "standard"
name = "windbarbs"


def call(
    xarray_obj,
    variable_name,
    area_def=None,
):
    """Coverage check routine for wind barb xarray object.

    Parameters
    ----------
    xarray_obj : xarray.Dataset
        xarray object containing variable "variable_name"
    variable_name : str
        variable name to check percent unmasked.

    Returns
    -------
    float
        Percent coverage of variable_name over area_def
    """
    if variable_name not in xarray_obj:
        raise KeyError(
            f"Variable {variable_name} did not exist. Can not calculate coverage."
        )

    from geoips.interfaces import interpolators

    interp_plugin = interpolators.get_plugin("interp_nearest")
    output_xarray = interp_plugin(
        area_def, xarray_obj, None, [variable_name], array_num=0
    )

    from geoips.data_manipulations.info import percent_unmasked

    return percent_unmasked(output_xarray[variable_name].to_masked_array())



"""Testing module for Pydantic FeatureAnnotatorPluginModel."""

from copy import deepcopy

import pytest

from geoips.pydantic.feature_annotators import FeatureAnnotatorPluginModel
from tests.unit_tests.pydantic.utils import (
    PathDict,
    load_test_cases,
    load_geoips_yaml_plugin,
    validate_bad_plugin,
    validate_good_plugin,
)

interface = "feature_annotators"

test_cases = load_test_cases(interface)
good_yaml = load_geoips_yaml_plugin(interface, "default_oldlace")


@pytest.fixture
def good_feature_annotator():
    """Return a consistent dictionary that is a valid GeoIPS feature annotator."""
    # Make the loading code only occur once, return a copy every time
    return PathDict(deepcopy(good_yaml))


@pytest.mark.parametrize(
    "good_feature_annotator",
    [pytest.param("good_feature_annotator", id="good_feature_annotator")],
    indirect=True,
)
def test_good_feature_annotator(good_feature_annotator):
    """Assert that a well formatted feature annotator plugin is valid.

    Parameters
    ----------
    good_feature_annotator: dict
        - A dictionary representing a valid feature annotator plugin.
    """
    validate_good_plugin(good_feature_annotator, FeatureAnnotatorPluginModel)


@pytest.mark.parametrize("test_tup", test_cases.values(), ids=list(test_cases.keys()))
def test_bad_feature_annotator_plugins(good_feature_annotator, test_tup):
    """Perform validation on feature_annotator plugins, including failing cases.

    Parameters
    ----------
    good_feature_annotator: dict
        - A dictionary representing a feature_annotator plugin that is valid.
    test_tup:
        - A tuple formatted (key, value, class, err_str), formatted (str, any, str, str)
          used to run and validate tests.
    """
    validate_bad_plugin(good_feature_annotator, test_tup, FeatureAnnotatorPluginModel)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for "Night_Vis_IR" product.

This algorithm expects two VIIRS channels (DNBRad and M16BT) for a RGB image
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "Night_Vis_IR"


def call(arrays):
    """Night_Vis_IR RGB product algorithm data manipulation steps.

    This algorithm expects DNBRad in reflectance and
    M16BT Brightness Temperatures in units of degrees Kelvin,
    and returns red green and blue gun arrays.

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data,
            in order of sensor "channels" list
        * Degrees Kelvin

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of qualitative RGBA image output

    Notes
    -----
    It will generate a product in daytime if we do not apply the daytime check.
    For now, it is for both day/night.

    We will decide whether this product is only for nighttime.
    If so, a daytime check will be required.

    We may focus only on nighttime product with moonlight after additional
    validation (TBD).
    """
    ch1 = arrays[0]  # Red gun: DNBRad
    ch2 = arrays[0]  # Green gun: DNBRad
    ch3 = arrays[1]  # Blue gun: M16BT (200-300K)

    val_max = ch1.max()

    if (
        val_max < 1.0e-3 and val_max >= 1.0e-8
    ):  # nighttime DNBRad is very small, normally aound 1.0e-8
        val_max = (
            0.05 * val_max
        )  # add a tuning factor 0.05 to the val_max with moonlight
    if val_max < 1.0e-8:
        val_max = (
            0.5 * val_max
        )  # add a tuning factor 0.5 to the val_max without moonlight

    data_range1 = [0, val_max]
    data_range2 = [200, 300]

    from geoips.data_manipulations.corrections import apply_data_range, apply_gamma

    red = ch1
    red = apply_data_range(
        red,
        min_val=data_range1[0],
        max_val=data_range1[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )  # need inverse option?
    red = apply_gamma(red, 1.0)

    grn = ch2
    grn = apply_data_range(
        grn,
        min_val=data_range1[0],
        max_val=data_range1[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )
    grn = apply_gamma(grn, 1.0)

    blu = ch3
    blu = apply_data_range(
        blu,
        min_val=data_range2[0],
        max_val=data_range2[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=True,
    )  # create image of deep clouds in blueish color
    # norm=True, inverse=False)    #create image of deep clouds in yellowish color
    blu = apply_gamma(blu, 1.0)

    from geoips.image_utils.mpl_utils import alpha_from_masked_arrays, rgba_from_arrays

    alp = alpha_from_masked_arrays([red, grn, blu])
    rgba = rgba_from_arrays(red, grn, blu, alp)

    return rgba



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS CLI "config" command.

Various configuration-based commands for setting up your geoips environment.
"""

from os import listdir, environ, remove
from os.path import abspath, join
import requests
import tempfile

from numpy import any
import tarfile
from tqdm import tqdm

from geoips.commandline.ancillary_info.test_data import test_dataset_dict
from geoips.commandline.geoips_command import GeoipsCommand, GeoipsExecutableCommand
from geoips.plugin_registry import PluginRegistry


class GeoipsConfigCreateRegistries(GeoipsExecutableCommand):
    """Config Command Class for creating plugin registries for plugin packages."""

    name = "create-registries"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the config-subparser for the Config Command."""
        self.parser.add_argument(
            "-s",
            "--save_type",
            default="json",
            type=str,
            choices=["json", "yaml"],
            help=(
                "The file format to save the registry as. Defaults to 'json', which is "
                "what's used by GeoIPS under the hood. For human readable output, you "
                "can provide the optional argument '-s yaml'."
            ),
        )

    def __call__(self, args):
        """Run the `geoips config create-registries -n <namespace> -s <save_type> -p <packages>` command.  # NOQA

        Parameters
        ----------
        args: Namespace()
            - The argument namespace to parse through
        """
        packages = args.packages
        namespace = args.namespace
        save_type = args.save_type
        plugin_registry = PluginRegistry(namespace)
        print(plugin_registry.create_registries(packages, save_type))


class GeoipsConfigDeleteRegistries(GeoipsExecutableCommand):
    """Config Command Class for deleting plugin registries for plugin packages."""

    name = "delete-registries"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the config-subparser for the Config Command."""
        pass

    def __call__(self, args):
        """Run the `geoips config delete-registries -n <namespace> -p <packages>` command.  # NOQA

        Parameters
        ----------
        args: Namespace()
            - The argument namespace to parse through
        """
        packages = args.packages
        namespace = args.namespace
        plugin_registry = PluginRegistry(namespace)
        print(plugin_registry.delete_registries(packages))


class GeoipsConfigInstall(GeoipsExecutableCommand):
    """Config Command Class for installing packages/data.

    Supports installation of packages and test data needed for testing and/or running
    your GeoIPS environment.
    """

    name = "install"
    command_classes = []

    @property
    def geoips_testdata_dir(self):
        """String path to GEOIPS_TESTDATA_DIR."""
        if not hasattr(self, "_geoips_testdata_dir"):
            self._geoips_testdata_dir = environ["GEOIPS_TESTDATA_DIR"]
        return self._geoips_testdata_dir

    def add_arguments(self):
        """Add arguments to the config-subparser for the Config Command."""
        self.parser.add_argument(
            "test_dataset_names",
            type=str.lower,
            nargs="+",
            choices=list(test_dataset_dict.keys()),
            help=(
                "Names of the GeoIPS test datasets to install. If 'all' is specified, "
                "GeoIPS will install all test datasets hosted on NextCloud. 'all' "
                "cannot be specified alongside other test dataset names."
            ),
        )
        self.parser.add_argument(
            "-o",
            "--outdir",
            type=str,
            default=self.geoips_testdata_dir,
            help="The full path to the directory you want to install this data to.",
        )

    def __call__(self, args):
        """Run the `geoips config install <test_dataset_names> -o <outdir>` command.

        Parameters
        ----------
        args: Namespace()
            - The argument namespace to parse through
        """
        test_dataset_names = args.test_dataset_names
        outdir = args.outdir

        if len(test_dataset_names) > 1 and "all" in test_dataset_names:
            self.parser.error(
                "Error: you cannot specify 'all' alongside other test dataset names. "
                "If 'all' is specified, that must be the only argument provided."
            )

        all_datasets = (
            True
            if len(test_dataset_names) == 1 and test_dataset_names[0] == "all"
            else False
        )

        install_dataset_names = (
            list(test_dataset_dict.keys()) if all_datasets else test_dataset_names
        )

        for test_dataset_name in install_dataset_names:
            test_dataset_url = test_dataset_dict[test_dataset_name]
            if any([test_dataset_name in fol for fol in listdir(outdir)]):
                print(
                    f"Test dataset '{test_dataset_name}' already exists under "
                    f"'{join(outdir, test_dataset_name)}*/'. See that "
                    "location for the contents of the test dataset."
                )
            else:
                print(
                    f"Installing {test_dataset_name} test dataset. This may take a "
                    "while..."
                )
                self.download_extract_test_data(test_dataset_url, outdir)
                out_str = (
                    f"Test dataset '{test_dataset_name}' has been installed under "
                    f"{outdir}/{test_dataset_name}/"
                )
                # Print the output of the command.
                print(out_str)

    def download_extract_test_data(self, url, download_dir):
        """Download the specified URL and write it to the corresponding download_dir.

        Will extract the data using tarfile and create an archive by bundling the
        associated files and directories together.

        Parameters
        ----------
        url: str
            - The url of the test dataset to download
        download_dir: str
            - The directory in which to download and extract the data into
        """
        resp = requests.get(url, stream=True, timeout=15)
        if resp.status_code == 200:

            total_size = int(resp.headers.get("Content-Length", 0))
            chunk_size = 1024 * 1024  # 1MB

            # Write the data to a temp file on disk. This is needed as files larger than
            # 1-2Gb might not fit in memory for some machines. Delete the temp file
            # after extraction has finished.
            with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
                # Progress bar setup
                progress = tqdm(
                    total=total_size,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                    desc="Downloading",
                )

                for chunk in resp.iter_content(chunk_size=chunk_size):
                    if chunk:
                        tmp_file.write(chunk)
                        progress.update(len(chunk))

                progress.close()
                tmp_file.flush()

                # Seek back to the start of the file for extraction
                tmp_file.seek(0)

            print("Beginning data extraction...")
            self.extract_data_cautiously(tmp_file.name, download_dir)
            # Delete the temp file!
            remove(tmp_file.name)
        else:
            self.parser.error(
                f"Error retrieving data from {url}; Status Code {resp.status_code}."
            )

    def extract_data_cautiously(self, filepath, download_dir):
        """Extract the GET Response cautiously and skip any dangerous members.

        Iterate through a Response and check that each member is not dangerous to
        extract to your machine. If it is, skip it.

        Where 'dangerous' is a filepath that is not part of 'download_dir'. File path
        maneuvering characters could be invoked ('../', ...), which we will not allow
        when downloading test data.

        Parameters
        ----------
        filepath: str
            - The path to the temporary file to extract from.
        download_dir: str
            - The directory in which to download and extract the data into
        """
        with tarfile.open(filepath, mode="r:gz") as tar:
            members = tar.getmembers()

            # Validate and extract each member of the archive
            with tqdm(
                total=len(members), unit="file", desc="Extracting", ncols=80
            ) as progress:
                for m in tar:
                    if not abspath(join(download_dir, m.name)).startswith(download_dir):
                        raise SystemExit("Found unsafe filepath in tar, exiting now.")
                    tar.extract(m, path=download_dir)
                    progress.update(1)


class GeoipsConfig(GeoipsCommand):
    """Config top-level command for configuring your GeoIPS environment."""

    name = "config"
    command_classes = [
        GeoipsConfigInstall,
        GeoipsConfigCreateRegistries,
        GeoipsConfigDeleteRegistries,
    ]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Feature Annotator interface module."""

from geoips.interfaces.base import BaseYamlInterface
from geoips.pydantic.feature_annotators import FeatureAnnotatorPluginModel


class FeatureAnnotatorsInterface(BaseYamlInterface):
    """Interface for feature annotator plugins."""

    name = "feature_annotators"
    validator = FeatureAnnotatorPluginModel


feature_annotators = FeatureAnnotatorsInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

# DO NOT EDIT
# managed by poetry-dynamic-versioning
__version__ = "1.16.0a0"
__version_tuple__ = (1, 16, "0a0")



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read derived surface winds from SAR, SMAP, SMOS, and AMSR netcdf data."""

# Python Standard Libraries
from datetime import datetime, timedelta
import logging
from os.path import basename

# Third-Party Libraries
import numpy as np
from netCDF4 import Dataset
import xarray

# GeoIPS imports
from geoips.xarray_utils.time import (
    get_min_from_xarray_time,
    get_max_from_xarray_time,
)

LOG = logging.getLogger(__name__)

MS_TO_KTS = 1.94384
DEG_TO_KM = 111.321

interface = "readers"
family = "standard"
name = "smos_winds_netcdf"
source_names = ["smos-spd"]


def read_smos_data(wind_xarray, fname):
    """Reformat SMOS xarray object appropriately.

    * variables: latitude, longitude, time, wind_speed_kts
    * attributes: source_name, platform_name, data_provider,
      interpolation_radius_of_influence
    """
    LOG.info("Reading SMOS data")

    # Attributes aren't set in the data files - use the file names to determine the
    # version.
    # 0 is good, 1 is fair, 2 is poor
    if "_110_" in fname:
        # Eliminate "poor" retrievals in the "old" version
        wind_xarray["wind_speed_kts"] = wind_xarray["wind_speed"].where(
            wind_xarray["quality_level"] < 2
        )[0, :, :]
    elif "_300_" in fname:
        # 2 eliminates everything anywhere close to land - initially we had kept
        # everything (< 3) because there were large data gaps without, but the data near
        # land really is poor quality so must remove.
        wind_xarray["wind_speed_kts"] = wind_xarray["wind_speed"].where(
            wind_xarray["quality_level"] < 2
        )[0, :, :]
    else:
        # Default to keeping at least fair quality, in case filenames change
        wind_xarray["wind_speed_kts"] = wind_xarray["wind_speed"].where(
            wind_xarray["quality_level"] < 2
        )[0, :, :]
    wind_xarray["wind_speed_kts"] = xarray.DataArray(
        data=np.flipud(wind_xarray.wind_speed_kts) * MS_TO_KTS,
        name="wind_speed_kts",
        coords=wind_xarray["wind_speed_kts"].coords,
    )

    # Set lat/lons appropriately
    # These are (1440x720)
    lats2d, lons2d = np.meshgrid(wind_xarray.lat, wind_xarray.lon)
    # Full dataset is 720x1440x2
    wind_xarray["latitude"] = xarray.DataArray(
        data=np.flipud(lats2d.transpose()),
        name="latitude",
        coords=wind_xarray["wind_speed_kts"].coords,
    )
    wind_xarray["longitude"] = xarray.DataArray(
        data=lons2d.transpose(),
        name="longitude",
        coords=wind_xarray["wind_speed_kts"].coords,
    )
    wind_xarray = wind_xarray.set_coords(["latitude", "longitude"])
    # timearray = np.zeros(wind_xarray.wind_speed_kts.shape).astype(int) +
    #                                                         wind_xarray.time.values[0]

    timearray = np.ma.masked_array(
        data=np.zeros(wind_xarray.wind_speed_kts.shape).astype(int)
        + wind_xarray.time.values[0],
        mask=True,
    )

    ncobj = Dataset(fname)
    basedt = np.datetime64(datetime.strptime("19900101", "%Y%m%d"))
    nctimearray = np.flipud(ncobj.variables["measurement_time"][...][0, :, :])
    timeinds = np.ma.where(nctimearray)
    # Check if there are any unmasked timeinds, if so update timearray
    if np.size(timeinds) > 0:
        timedata = nctimearray[timeinds].data.tolist()
        timevals = np.ma.masked_array(
            [basedt + np.timedelta64(timedelta(days=xx)) for xx in timedata]
        )
        timearray[timeinds] = timevals
    # Otherwise set timearray as unmasked values
    else:
        timearray = timearray.data
    wind_xarray["time"] = xarray.DataArray(
        data=timearray, name="time", coords=wind_xarray["wind_speed_kts"].coords
    )
    return {"WINDSPEED": wind_xarray}


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read SMOS derived winds from netcdf data.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    ingested = []
    for fname in fnames:
        wind_xarray = xarray.open_dataset(str(fname))
        # Set attributes appropriately
        wind_xarray.attrs["source_file_names"] = [basename(fname)]
        wind_xarray.attrs["minimum_coverage"] = 20
        wind_xarray.attrs["source_name"] = "smos-spd"
        wind_xarray.attrs["platform_name"] = "smos"
        wind_xarray.attrs["data_provider"] = "esa"
        wind_xarray.attrs["interpolation_radius_of_influence"] = 25000
        # wind_xarray.attrs['sample_distance_km'] = DEG_TO_KM / 4
        wind_xarray.attrs["sample_distance_km"] = 25.0

        # Checking if the wind_xarray.time is valid
        if (
            not isinstance(wind_xarray.time.values[0], np.datetime64)
            and wind_xarray.time.values[0].year > 3000
        ):

            cov_start = datetime.strptime(
                wind_xarray.time_coverage_start, "%Y-%m-%dT%H:%M:%S Z"
            )
            cov_end = datetime.strptime(
                wind_xarray.time_coverage_end, "%Y-%m-%dT%H:%M:%S Z"
            )
            time = cov_start + timedelta(
                seconds=(cov_end - cov_start).total_seconds() / 2
            )
            time_attrs = wind_xarray.time.attrs
            wind_xarray["time"] = np.array([time], dtype=np.datetime64)
            wind_xarray["time"].attrs = time_attrs

        LOG.info("Read data from %s", fname)

        # SMOS time is not read in correctly natively with xarray - must pass fname so
        # we can get time information directly from netCDF4.Dataset open
        ingested += [read_smos_data(wind_xarray, fname)]

    final_wind_xarrays = {}
    # SMOS is a global
    for wind_xarrays in ingested:
        if "WINDSPEED" not in final_wind_xarrays:
            final_wind_xarrays["WINDSPEED"] = wind_xarrays["WINDSPEED"]
        else:
            curr_xobj = wind_xarrays["WINDSPEED"]
            final_xobj = final_wind_xarrays["WINDSPEED"]
            windspd = xarray.where(
                curr_xobj["wind_speed_kts"] >= 0,
                curr_xobj["wind_speed_kts"],
                final_xobj["wind_speed_kts"],
            )
            timestmp = xarray.where(
                curr_xobj["wind_speed_kts"] >= 0,
                curr_xobj["time"],
                final_xobj["time"],
            )
            # for varname in curr_xobj.variables:
            #     if curr_xobj["wind_speed_kts"].shape == curr_xobj[varname].shape:
            #         final_xobj[varname] = xarray.where(
            final_wind_xarrays["WINDSPEED"] = xarray.concat(
                [final_wind_xarrays["WINDSPEED"], wind_xarrays["WINDSPEED"]],
                dim="time",
            )
            # This is lame.  Interpolation did not immediately handle
            # 6x721x1442 array
            final_wind_xarrays["WINDSPEED"]["time"] = timestmp
            final_wind_xarrays["WINDSPEED"]["wind_speed_kts"] = windspd

    for wind_xarray in final_wind_xarrays.values():
        LOG.info("Setting standard metadata")
        wind_xarray.attrs["start_datetime"] = get_min_from_xarray_time(
            wind_xarray, "time"
        )
        wind_xarray.attrs["end_datetime"] = get_max_from_xarray_time(
            wind_xarray, "time"
        )

        if "wind_speed_kts" in wind_xarray.variables:
            # These text files store wind speeds natively in kts
            wind_xarray["wind_speed_kts"].attrs["units"] = "kts"

        LOG.info(
            "Read data %s start_dt %s source %s platform %s data_provider %s roi "
            "%s native resolution",
            wind_xarray.attrs["start_datetime"],
            wind_xarray.attrs["source_name"],
            wind_xarray.attrs["platform_name"],
            wind_xarray.attrs["data_provider"],
            wind_xarray.attrs["interpolation_radius_of_influence"],
            wind_xarray.attrs["sample_distance_km"],
        )

    final_wind_xarrays["METADATA"] = wind_xarray[[]]
    if (
        final_wind_xarrays["METADATA"].start_datetime
        == final_wind_xarrays["METADATA"].end_datetime
    ):
        # Use alternate attributes to set start and end datetime

        final_wind_xarrays["METADATA"].attrs["start_datetime"] = datetime.strptime(
            wind_xarray.time_coverage_start, "%Y-%m-%dT%H:%M:%S Z"
        )
        final_wind_xarrays["METADATA"].attrs["end_datetime"] = datetime.strptime(
            wind_xarray.time_coverage_end, "%Y-%m-%dT%H:%M:%S Z"
        )
    return final_wind_xarrays



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module containing matplotlib information for RGB or RGBA imagery."""

import logging

LOG = logging.getLogger(__name__)

interface = "colormappers"
family = "matplotlib"
name = "cmap_rgb"


def call():
    """For rgb imagery, we require no color information.

    colormap is entirely specified by the RGB(A) arrays, so no specific
    matplotlib color information required.

    Parameters
    ----------
        No arguments

    Returns
    -------
    mpl_colors_info : dict
        * Specifies matplotlib Colors parameters for use in both plotting
          and colorbar generation
        * For RGBA arrays, all fields are "None"
    """
    mpl_colors_info = {
        "cmap": None,
        "norm": None,
        "cbar_ticks": None,
        "cbar_tick_labels": None,
        "cbar_label": None,
        "boundaries": None,
        "cbar_spacing": "proportional",
        "colorbar": False,
    }
    return mpl_colors_info



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""General high level utilities for geoips processing."""

import argparse
import inspect
import os
from copy import deepcopy
from shutil import get_terminal_size
import json
from pathlib import Path
import logging
from importlib import metadata, resources, import_module

from tabulate import tabulate
import numpy as np

from geoips.errors import PluginRegistryError, PluginPackageNotFoundError
from geoips.filenames.base_paths import PATHS as geoips_paths

LOG = logging.getLogger(__name__)


def get_interface_module(namespace):
    """Retrieve the interface module from a given namespace.

    Since this function uses the first portion of a namespace, I.e.
    (geoips.plugin_packages --> geoips), only interfaces implemented in that exact
    package will be recognized in that namespace. This means that if developers or users
    implement new interfaces in the 'geoips.plugin_packages' namespace, they still won't
    be recognized or available for use in GeoIPS.

    To get new interfaces to work, you must implement them in a separate namespace that
    is named the same as the plugin package that implemented them. I.e.
    (splunk.plugin_packages --> splunk). Any other package that falls under that
    namespace can make use of these interfaces, as well as GeoIPS' interfaces.

    Parameters
    ----------
    namespace: str
        - The namespace containg the requested interface module. I.e.
          'geoips.plugin_packages'.
    """
    package_name = namespace.split(".")[0]
    return import_module(package_name).interfaces


def remove_unsupported_kwargs(module, requested_kwargs):
    """Remove unsupported keyword arguments."""
    module_args = set(inspect.signature(module).parameters.keys())
    unsupported = list(set(requested_kwargs.keys()).difference(module_args))
    if "kwargs" not in module_args:
        for key in unsupported:
            LOG.warning("REMOVING UNSUPPORTED %s key %s", module, key)
            requested_kwargs.pop(key)
    return requested_kwargs


def find_ascii_palette(name):
    """Find ASCII palette named "name".

    Search the plugins/txt/ascii_palettes directory for ASCII palettes to use
    as colormaps.
    """
    all_plugins = find_all_txt_plugins("txt/ascii_palettes")

    for plugin in all_plugins:
        if name == os.path.splitext(os.path.basename(plugin))[0]:
            return plugin
    raise ValueError(f"Non-existent txt plugin: {name}")


def find_all_txt_plugins(subdir=""):
    """Find all txt plugins in registered plugin packages.

    Search the ``plugins`` directory of each registered plugin package for files ending
    in ``.txt``. Return list of files
    """
    # Load all entry points for plugin packages
    plugin_packages = metadata.entry_points(group="geoips.plugin_packages")

    # Loop over the plugin packages and load all of their yaml plugins
    txt_files = []
    for pkg in plugin_packages:
        pkg_plugin_path = resources.files(pkg.value) / "plugins" / subdir
        txt_files += pkg_plugin_path.rglob("*.txt")

    return txt_files


def load_all_yaml_plugins():
    """Find all YAML plugins in registered plugin packages.

    Search the ``plugins`` directory of each registered plugin package for files ending
    in ``.yaml``. Read each plugin file
    """
    # Load all entry points for plugin packages
    import json

    plugin_packages = metadata.entry_points(group="geoips.plugin_packages")
    yaml_plugins = {}
    for pkg in plugin_packages:
        pkg_plug_path = str(resources.files(pkg.value) / "registered_plugins")
        if not os.path.exists(pkg_plug_path):
            raise PluginRegistryError(
                f"Plugin registry {pkg_plug_path} did not exist, "
                "please run 'create_plugin_registries'"
            )
        # This will include all plugins, including schemas, yaml_based,
        # and module_based plugins.
        registered_plugins = json.load(open(pkg_plug_path, "r"))
        # Only pull the "yaml_based" plugins here.
        try:
            for interface in registered_plugins["yaml_based"]:
                if interface not in yaml_plugins:
                    yaml_plugins[interface] = registered_plugins["yaml_based"][
                        interface
                    ]
                else:
                    merge_nested_dicts(
                        yaml_plugins[interface],
                        registered_plugins["yaml_based"][interface],
                    )
        except TypeError:
            raise PluginRegistryError(f"Failed reading {pkg_plug_path}.")
    return yaml_plugins


def copy_standard_metadata(orig_xarray, dest_xarray, extra_attrs=None, force=True):
    """Copy standard metadata from orig_xarray to dest_xarray.

    Parameters
    ----------
    orig_xarray : xarray.Dataset
        Original xarray to copy attributes from
    dest_xarray : xarray.Dataset
        Destination xarray to copy attributes to
    extra_attrs : list of str, optional
        Additional attributes to copy, beyond the standard metadata, by default None
    force : bool, optional
        If force is True, overwrite existing attributes, by default True

    Returns
    -------
    xarray.Dataset
        dest_xarray with standard metadata copied in place from orig_xarray.
    """
    attrs = [
        "start_datetime",
        "end_datetime",
        "platform_name",
        "source_name",
        "minimum_coverage",
        "data_provider",
        "granule_minutes",
        "original_source_filenames",
        "source_file_names",
        "sample_distance_km",
        "interpolation_radius_of_influence",
        "area_definition",
    ]
    if extra_attrs is not None:
        attrs += extra_attrs

    for attr in attrs:
        if force and attr in orig_xarray.attrs.keys():
            dest_xarray.attrs[attr] = orig_xarray.attrs[attr]
        elif (
            not force
            and attr in orig_xarray.attrs.keys()
            and attr not in dest_xarray.attrs.keys()
        ):
            dest_xarray.attrs[attr] = orig_xarray.attrs[attr]


def deprecation(message):
    """Print a deprecation warning during runtime."""
    import warnings

    warnings.warn(message, DeprecationWarning, stacklevel=2)
    LOG.warning("DeprecationWarning: %s", message)


def output_process_times(process_datetimes, num_jobs=None, job_str="GeoIPS 2"):
    """Calculate and print the process times from the process_datetimes dictionary.

    Parameters
    ----------
    process_datetimes : dict
        dictionary formatted as follows:

        * ``process_datetimes['overall_start']`` - overall start datetime of
          the entire script
        * ``process_datetimes['overall_end']`` - overall end datetime of the
          entire script
        * ``process_datetimes[process_name]['start']`` - start time of an
          individual process
        * ``process_datetimes[process_name]['end']`` - end time of an
          individual process
    """
    if "overall_end" in process_datetimes and "overall_start" in process_datetimes:
        LOG.info(
            "Total Time %s: %s Num jobs: %s",
            process_datetimes["overall_end"] - process_datetimes["overall_start"],
            num_jobs,
            job_str,
        )
    for process_name in process_datetimes.keys():
        if process_name in ["overall_start", "overall_end"]:
            continue
        if "end" in process_datetimes[process_name]:
            LOG.info(
                "    SUCCESS Process Time %s: %-20s: %s",
                job_str,
                process_name,
                process_datetimes[process_name]["end"]
                - process_datetimes[process_name]["start"],
            )
        elif "fail" in process_datetimes[process_name]:
            LOG.info(
                "    FAILED  Process Time %s: %-20s: %s",
                job_str,
                process_name,
                process_datetimes[process_name]["fail"]
                - process_datetimes[process_name]["start"],
            )
        else:
            LOG.info("    MISSING Process Time %s: %s", job_str, process_name)


def order_paths_from_least_to_most_specific(paths):
    """
    Orders a list of filesystem paths from least to most specific.

    This function takes a list of filesystem paths and returns a new list of paths
    ordered from the least specific (higher-level directories) to the most specific
    (subdirectories and files). It expands environmental variables in paths.

    Parameters
    ----------
    paths : list of str or pathlib.Path
        A list of filesystem paths to be ordered.

    Returns
    -------
    list of pathlib.Path
        A list of filesystem paths ordered from least to most specific.

    Examples
    --------
    >>> paths = [
    ...     '/home/user/docs/',
    ...     '/home/user/images/',
    ...     '/home/user/',
    ...     '/home/user/images/photo.jpg'
    ...     '/home/user/docs/report.txt',
    ... ]
    >>> order_paths_from_least_to_most_specific(paths)
    [PosixPath('/home/user/'),
     PosixPath('/home/user/docs/'),
     PosixPath('/home/user/images/'),
     PosixPath('/home/user/docs/report.txt'),
     PosixPath('/home/user/images/photo.jpg')]

    """
    if not paths:
        return []
    ordered_paths = []
    unordered_paths = []
    paths = [Path(os.path.expandvars(p)) for p in paths]
    for i, path in enumerate(paths):
        other_paths = paths[:i] + paths[i + 1 :]
        if all([path not in other_path.parents for other_path in other_paths]):
            # path not in other paths, least specific already
            ordered_paths.append(path)
        else:
            # path in other path, needs more sorting
            unordered_paths.append(path)
    return ordered_paths + order_paths_from_least_to_most_specific(unordered_paths)


def replace_geoips_paths_in_list(
    replace_list, replace_paths=None, base_paths=None, curly_braces=False
):
    """
    Replace GeoIPS paths with geoips settings in elements of a list.

    This function iterates over each element in the provided `replace_list`,
    attempting to replace GeoIPS paths within each element using the
    `replace_geoips_paths` function. If an element raises a `TypeError`
    when cast to a pathlib path, it is skipped.

    Parameters
    ----------
    replace_list : list
        A list of elements to process. Elements can be of any type, but only those that
        are Path-like will be processed.
    replace_paths : dict, optional
        Passed to replace_geoips_paths
    base_paths : dict, optional
        Passed to replace_geoips_paths
    curly_braces : bool, optional
        Passed to replace_geoips_paths

    Returns
    -------
    list
        A new list containing the elements with GeoIPS paths replaced where possible.
        Elements that could not be processed are included unchanged.

    Examples
    --------
    >>> replace_geoips_paths_in_list(['/home/geoips/data/project',
    ... 'no_replacement_here'])
    ['$GEOIPS_DATA_DIR/project', 'no_replacement_here']

    See Also
    --------
    replace_geoips_paths : Function used to replace GeoIPS paths in individual elements.
    """
    new_list = []
    # Go through each element in the list
    for val in replace_list:
        try:
            new_list.append(
                replace_geoips_paths(
                    val,
                    replace_paths=replace_paths,
                    base_paths=base_paths,
                    curly_braces=curly_braces,
                )
            )
        except TypeError:
            # Otherwise, just put the current element back
            new_list.append(val)
            continue
    return new_list


def replace_geoips_paths_in_dict(
    replace_dict, replace_paths=None, base_paths=None, curly_braces=False
):
    """Replace geoips paths in every path-based element within a dictionary."""
    dump_dict = replace_dict.copy()
    for key in replace_dict:
        # If this is a string, and it contains "/", replace the geoips paths.
        if isinstance(replace_dict[key], str) and "/" in replace_dict[key]:
            dump_dict[key] = replace_geoips_paths(replace_dict[key])
        # If this is a list, go through each element and replace geoips paths if
        # applicable.
        if isinstance(replace_dict[key], list):
            dump_dict[key] = replace_geoips_paths_in_list(
                replace_dict[key], replace_paths, base_paths, curly_braces
            )
    return dump_dict


def replace_geoips_paths(
    path,
    replace_paths=None,
    base_paths=None,
    curly_braces=False,
):
    """Replace specified sub-paths in path with related environment variable names.

    This function replaces paths in the provided path with their corresponding
    environment variable names. This is useful for generating output paths
    or metadata that are independent of specific installation directories.

    For example, it can replace:

    - ``'/home/user/geoproc/geoips_packages'`` with ``'$GEOIPS_PACKAGES_DIR'``
    - ``'/home/user/geoproc/geoips_outdirs'`` with ``'$GEOIPS_OUTDIRS'``
    - ``'/home/user/geoproc'`` with ``'$GEOIPS_BASEDIR'``

    Parameters
    ----------
    path : str or pathlib.Path
        The path in which to replace base paths.
    replace_paths : list of str, optional
        A list of environment variable names whose corresponding paths should be
        replaced in `path`.
        If `None`, defaults to:

        ``['$GEOIPS_OUTDIRS', '$GEOIPS_PACKAGES_DIR', '$GEOIPS_TESTDATA_DIR',
        '$GEOIPS_DEPENDENCIES_DIR', '$GEOIPS_BASEDIR']``
    base_paths : dict, optional
        A dictionary mapping environment variable names to their corresponding base
        paths.  If `None`, defaults to `geoips.filenames.base_paths.PATH`.
    curly_braces : bool, default=False
        If `True`, includes curly braces in the environment variables
        (e.g., ``'${GEOIPS_BASEDIR}'``),
        otherwise excludes them (e.g., ``'$GEOIPS_BASEDIR'``).

    Returns
    -------
    str
        The path with specified base paths replaced with environment variable names.

    Notes
    -----
    The function iterates over the provided `replace_paths` in reverse order
    (from most specific to least specific) and replaces the first matching base
    path in the given `path` with the corresponding environment variable name.

    Examples
    --------
    >>> path = '/home/user/geoproc/geoips_packages/module/file.py'
    >>> base_paths = {
    ...     'GEOIPS_PACKAGES_DIR': '/home/user/geoproc/geoips_packages',
    ...     'GEOIPS_BASEDIR': '/home/user/geoproc'
    ... }
    >>> replace_geoips_paths(path, base_paths=base_paths)
    '$GEOIPS_PACKAGES_DIR/module/file.py'
    """
    # Allow multiple sets of base_path replacements

    if base_paths is None:
        base_paths = geoips_paths

    # These are the environment variables that are specified in base_paths.py.
    # Eventually we will want to pull these directly from the environment config,
    # for now explicitly list env vars here.
    if replace_paths is None:
        replace_env_vars = [
            "$TCWWW",
            "$PRIVATEWWW",
            "$PUBLICWWW",
            "$GEOTIFF_IMAGERY_PATH",
            "$ANNOTATED_IMAGERY_PATH",
            "$CLEAN_IMAGERY_PATH",
            "$GEOIPS_OUTDIRS",
            "$GEOIPS_PACKAGES_DIR",
            "$GEOIPS_TESTDATA_DIR",
            "$GEOIPS_DEPENDENCIES_DIR",
            "$GEOIPS_BASEDIR",
        ]

    paths_to_be_replaced = [Path(os.path.expandvars(p)) for p in replace_env_vars]
    ordered_path_envvar_dict = {
        replace_env_vars[paths_to_be_replaced.index(replace_path)]: replace_path
        for replace_path in order_paths_from_least_to_most_specific(
            paths_to_be_replaced
        )
    }

    # Replace with specified file system -> URL mapping
    # for paths in base_paths:
    #    for key in paths.keys():
    #        if f"{key}_URL" in paths:
    #            fname = fname.replace(paths[key], paths[f"{key}_URL"])

    path = Path(os.path.expandvars(path))

    # Replace full paths with environment variables
    for env_var, replace_path in ordered_path_envvar_dict.items():
        if replace_path in path.parents:
            env_var = env_var.replace("$", "")
            return str(path).replace(
                str(replace_path),
                f"${{{env_var}}}" if curly_braces else f"${env_var}",
            )
    return str(path)


def get_required_geoips_xarray_attrs():
    """Interface deprecated v2.0."""
    required_xarray_attrs = [
        "source_name",
        "platform_name",
        "data_provider",
        "start_datetime",
        "end_datetime",
    ]
    return required_xarray_attrs


def merge_nested_dicts(dest, src, in_place=True, replace=False):
    """Perform an in-place merge of src into dest.

    Performs an in-place merge of src into dest while preserving any values that already
    exist in dest.
    """
    if not in_place:
        final_dest = deepcopy(dest)
    else:
        final_dest = dest

    # NOTE: this is a top-level field, where if you set
    # product_spec_override:
    #   replace: true
    # It will automatically replace ALL fields found in
    # the original product spec and also found in the
    # override with what is specified in the override
    # in its entirety, without merging.  This is not
    # terribly useful overall - we probably want this
    # sort of capability in the end, but more flexible
    # and able to be applied to only specific fields,
    # etc.  This is a brute force method to at least
    # allow overriding entire fields.
    if replace:
        for key in final_dest:
            if key in src:
                final_dest[key] = src[key]
        return final_dest

    try:
        final_dest.update(src | final_dest)
    except (AttributeError, TypeError):
        return
    try:
        for key, val in final_dest.items():
            try:
                merge_nested_dicts(final_dest[key], src[key])
            except KeyError:
                pass
    except AttributeError:
        raise
    if not in_place:
        return final_dest


def expose_geoips_commands(pkg_name=None, _test_log=None):
    """Expose a list of commands that operate in the GeoIPS environment.

    Where, these commands are defined under 'pyproject.toml:[tool.poetry.scripts]',
    or 'pyproject.toml:[project.entry-points.console_scripts]'

    Parameters
    ----------
    pkg_name: str (default = None)
        - The name of the GeoIPS Plugin Package whose command's will be exposed.
        - If None, assume this was called via the commandline and retrieve package_name
          via that manner. Otherwise use the supplied package_name.
    _test_log: logging.Logger (default = None)
        - If provided, use this logger instead. This is added as an optional argument
          so we can check the output of this command for our Unit Tests.
    """
    pkg_name, log = _get_pkg_name_and_logger(pkg_name, _test_log)
    # Get a list of console_script entrypoints specific to the provided package
    eps = list(
        filter(
            lambda ep: pkg_name in ep.value,
            metadata.entry_points().select(group="console_scripts"),
        )
    )
    log.interactive("-" * len(f"Available {pkg_name.title()} Commands"))
    log.interactive(f"Available {pkg_name.title()} Commands")
    log.interactive("-" * len(f"Available {pkg_name.title()} Commands"))
    if eps:
        table_data = [[ep.name, ep.value] for ep in eps]
        # Log the commands found in a tabular fashion
        log.interactive(
            tabulate(
                table_data,
                headers=["Command Name", "Command Path"],
                tablefmt="rounded_grid",
                maxcolwidths=get_terminal_size().columns // 2,
            )
        )
        # Otherwise let the user know that there were not commands found for this
        # package.
    else:
        log.interactive(f"No '{pkg_name.title()}' Commands were found.")


def _get_pkg_name_and_logger(pkg_name, provided_log):
    """Return the corresponding package name and logger for exposing package commands.

    If pkg_name is None, retrieve pkg_name from the commandline arguments
    (either -p <package_name> or default 'geoips'). If log is None, set log to the LOG
    attribute found in this module.

    If either variable isn't None, use what's provided instead. This is used for
    unit testing primarily.

    Parameters
    ----------
    pkg_name: str
        - The name of the GeoIPS Plugin Package whose command's will be exposed.
        - If None, assume this was called via the commandline and retrieve package_name
          via that manner. Otherwise use the supplied package_name.
        - If supplied, pkg_name must be one of pip installed 'geoips.plugin_packages'.
          ie. 'recenter_tc', 'geoips_clavrx', 'data_fusion', <your_custom_pkg>, ...
    provided_log: logging.Logger
        - If None, retrieve the LOG attribute from this module, otherwise use the
          provided logger so we can check the output of this function for unit tests.

    Returns
    -------
    pkg_name, log: str, logging.Logger
        - The name of the package to retrieve commands from and the logger used to
          output it.
    """
    plugin_packages = [
        str(ep.value) for ep in metadata.entry_points(group="geoips.plugin_packages")
    ]
    if provided_log:
        log = provided_log
    else:
        log = LOG
    if pkg_name is None:
        # This function was called via the command line or None was passed.
        argparser = argparse.ArgumentParser("expose command")
        argparser.add_argument(
            "--package_name",
            "-p",
            type=str.lower,
            default="geoips",
            choices=plugin_packages,
            help="GeoIPS Plugin package to expose.",
        )
        ARGS = argparser.parse_args()
        pkg_name = ARGS.package_name
    else:
        # This function was called via python
        if pkg_name not in plugin_packages:
            raise PluginPackageNotFoundError(
                f"No such package named '{pkg_name}' found. Make sure that package is "
                "installed with a package manager such as pip."
            )
    return pkg_name, log


def is_editable(package_name):
    """Return whether or not 'package_name' has been installed in editable mode.

    Where editable mode is a local package installed via 'pip install -e <path_to_pkg>
    and non-editable mode is a local package installed via 'pip install <pact_to_pkg>.

    If the package under package_name doesn't exist, raise a ValueError reporting that.

    Parameters
    ----------
    package_name: str
        - The name of the pip installed local package. (ie. "geoips", "recenter_tc", ..)

    Returns
    -------
    editable: bool
        - The truth value as to whether or not the package was installed in editable
          mode.
    """
    plugin_package_names = [
        ep.value for ep in metadata.entry_points(group="geoips.plugin_packages")
    ]
    if package_name not in plugin_package_names:
        raise ValueError(
            f"Package '{package_name}' is not an installed package. Please install it "
            "before running this command; ie via 'pip install <path_to_package_name>' "
            "optionally with the '-e' flag. You can also use another package manager."
        )
    dist_info = metadata.distribution(package_name).read_text("direct_url.json")
    # If dist_info is None, package was not installed from source and it was installed
    # from a pre-built wheel. Therefore it is not in editable mode.
    if dist_info:
        # If dist_info is not None, that means we retrieved metadata about the installed
        # package. Check to see if it's in editable mode or not.
        json_dist = json.loads(dist_info)
        if (
            "dir_info" in json_dist.keys()
            and "editable" in json_dist["dir_info"].keys()
            and json_dist["dir_info"]["editable"]
        ):
            # If the 'editable' key exists and is True
            return True
    # Package is installed in non-editable mode
    return False


def get_numpy_seeded_random_generator():
    """
    Get a NumPy random generator seeded with a fixed value.

    Returns
    -------
    numpy.random.Generator
        A NumPy random generator initialized with a fixed seed value of 42.

    Notes
    -----
    This function returns a seeded random generator using NumPy's `default_rng`
    function with a fixed seed of 42. Using a fixed seed ensures that the
    random numbers generated by this generator will be reproducible across different
    runs. 42 was chosen because it is the answer to the universe and all things;
    that is to say: it's selection is inconsequential and any other number
    could have been chosen.

    Examples
    --------
    >>> predictable_random = get_numpy_seeded_random_generator()
    >>> predictable_random.integers(0, 10, size=5)
    array([6, 3, 7, 4, 6])  # Example output, will be the same every time

    """
    return np.random.default_rng(seed=42)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read derived surface winds from BYU ASCAT UHR NetCDF data."""

# Python Standard Libraries
from datetime import datetime, timedelta
import logging
import os

# Third-Party Libraries
import numpy
import xarray

# GeoIPS imports
from geoips.xarray_utils.time import (
    get_min_from_xarray_time,
    get_max_from_xarray_time,
)

LOG = logging.getLogger(__name__)

MS_TO_KTS = 1.94384
DEG_TO_KM = 111.321

interface = "readers"
family = "standard"
name = "ascat_uhr_netcdf"
source_names = ["ascatuhr"]


def read_byu_data(wind_xarray, fname):
    """Reformat ascat xarray object appropriately.

    * variables: latitude, longitude, time,
      wind_speed_kts, wind_dir_deg_met
    * attributes: source_name, platform_name, data_provider,
      interpolation_radius_of_influence
    """
    if "L2B_filename" in wind_xarray.attrs and "metopa" in wind_xarray.L2B_filename:
        wind_xarray.attrs["source_name"] = "ascatuhr"
        wind_xarray.attrs["platform_name"] = "metop-a"
    elif "L2B_filename" in wind_xarray.attrs and "metopb" in wind_xarray.L2B_filename:
        wind_xarray.attrs["source_name"] = "ascatuhr"
        wind_xarray.attrs["platform_name"] = "metop-b"
    elif "L2B_filename" in wind_xarray.attrs and "metopc" in wind_xarray.L2B_filename:
        wind_xarray.attrs["source_name"] = "ascatuhr"
        wind_xarray.attrs["platform_name"] = "metop-c"
    elif "SZF_filenames" in wind_xarray.attrs and "M02" in wind_xarray.SZF_filenames:
        wind_xarray.attrs["source_name"] = "ascatuhr"
        wind_xarray.attrs["platform_name"] = "metop-a"
    elif "SZF_filenames" in wind_xarray.attrs and "M01" in wind_xarray.SZF_filenames:
        wind_xarray.attrs["source_name"] = "ascatuhr"
        wind_xarray.attrs["platform_name"] = "metop-b"
    elif "SZF_filenames" in wind_xarray.attrs and "M03" in wind_xarray.SZF_filenames:
        wind_xarray.attrs["source_name"] = "ascatuhr"
        wind_xarray.attrs["platform_name"] = "metop-c"
    elif (
        "source_file_names" in wind_xarray.attrs
        and "_C_" in wind_xarray.attrs["source_file_names"][0]
    ):
        wind_xarray.attrs["source_name"] = "ascatuhr"
        wind_xarray.attrs["platform_name"] = "metop-c"
    elif (
        "source_file_names" in wind_xarray.attrs
        and "_B_" in wind_xarray.attrs["source_file_names"][0]
    ):
        wind_xarray.attrs["source_name"] = "ascatuhr"
        wind_xarray.attrs["platform_name"] = "metop-b"

    new_file = False
    try:
        # Store the storm names lower case - only reference to it is in the filename..
        storm_name = (
            os.path.basename(wind_xarray.source_file_names[0]).split("_")[3].lower()
        )
        expected_yymmdd = os.path.basename(wind_xarray.source_file_names[0]).split("_")[
            4
        ]
        expected_hhmn = os.path.basename(
            wind_xarray.source_file_names[0]
            .replace(".WRave3.nc", "")
            .replace(".avewr.nc", "")
        ).split("_")[5]
        wind_xarray.attrs["expected_synoptic_time"] = datetime.strptime(
            expected_yymmdd + expected_hhmn, "%y%m%d%H%M"
        )
        wind_xarray.attrs["storms_with_coverage"] = [storm_name]
        new_file = False
    except IndexError:
        # New filenames as of 20220826, no time, only YYYYMMDD, orbit num, and A or D
        # for ascending or descending
        # MUIFA_20220911_19947_C_D-product.nc
        # Store the storm names lower case - only reference to it is in the filename..
        storm_name = (
            os.path.basename(wind_xarray.source_file_names[0]).split("_")[0].lower()
        )
        wind_xarray.attrs["storms_with_coverage"] = [storm_name]
        new_file = True
    except ValueError:
        # New filenames as of 20230906, only YYYYMMDD, orbit num,
        # and A or D for ascending or descending. Example file name:
        # LEE_20230912_25146_C_A-cmod5h-scaled_v2.nc
        # Store the storm names lower case - only reference to it is in the filename.
        storm_name = (
            os.path.basename(wind_xarray.source_file_names[0]).split("_")[0].lower()
        )
        wind_xarray.attrs["storms_with_coverage"] = [storm_name]
        new_file = True

    dsname = "DATA"
    if "wspeeds" in wind_xarray.variables:
        wind_xarray["wind_speed_kts"] = xarray.where(
            wind_xarray.ambiguity_select == 1, wind_xarray.wspeeds[:, :, 0], numpy.nan
        )
        wind_xarray["wind_speed_kts"] = xarray.where(
            wind_xarray.ambiguity_select == 2,
            wind_xarray.wspeeds[:, :, 1],
            wind_xarray.wind_speed_kts,
        )
        wind_xarray["wind_speed_kts"] = xarray.where(
            wind_xarray.ambiguity_select == 3,
            wind_xarray.wspeeds[:, :, 2],
            wind_xarray.wind_speed_kts,
        )
        wind_xarray["wind_speed_kts"] = xarray.where(
            wind_xarray.ambiguity_select == 4,
            wind_xarray.wspeeds[:, :, 3],
            wind_xarray.wind_speed_kts,
        )
        wind_xarray["wind_speed_kts"] = wind_xarray["wind_speed_kts"] * MS_TO_KTS

        wind_xarray["wind_dir_deg_met"] = xarray.where(
            wind_xarray.ambiguity_select == 1, wind_xarray.wdirs[:, :, 0], numpy.nan
        )
        wind_xarray["wind_dir_deg_met"] = xarray.where(
            wind_xarray.ambiguity_select == 2,
            wind_xarray.wdirs[:, :, 1],
            wind_xarray.wind_dir_deg_met,
        )
        wind_xarray["wind_dir_deg_met"] = xarray.where(
            wind_xarray.ambiguity_select == 3,
            wind_xarray.wdirs[:, :, 2],
            wind_xarray.wind_dir_deg_met,
        )
        wind_xarray["wind_dir_deg_met"] = xarray.where(
            wind_xarray.ambiguity_select == 4,
            wind_xarray.wdirs[:, :, 3],
            wind_xarray.wind_dir_deg_met,
        )
        # Set wind_speed_kts appropriately
        wind_xarray["wind_speed_kts"].attrs = wind_xarray["wspeeds"].attrs.copy()
        wind_xarray["wind_speed_kts"].attrs["units"] = "kts"
        if wind_xarray["wspeeds"].attrs.get("long_name"):
            wind_xarray["wind_speed_kts"].attrs["long_name"] = (
                wind_xarray["wspeeds"]
                .attrs["long_name"]
                .replace("ambiguities", "ambiguity selection")
            )

        wind_xarray["wind_dir_deg_met"].attrs = wind_xarray["wdirs"].attrs.copy()

        # Set lat/lons/time appropriately
        wind_xarray = wind_xarray.rename(
            {
                "wspeeds": "wind_speed_ambiguities_kts",
                "wdirs": "wind_dir_ambiguities_deg_met",
            }
        )
        wind_xarray["wind_speed_ambiguities_kts"] = (
            wind_xarray["wind_speed_ambiguities_kts"] * MS_TO_KTS
        )

        if "lat" in wind_xarray.variables:
            wind_xarray = wind_xarray.rename({"lat": "latitude", "lon": "longitude"})
        wind_xarray["latitude"] = xarray.where(
            wind_xarray.ambiguity_select == 0, numpy.nan, wind_xarray.latitude
        )
        if not new_file:
            # Old files were 0 to 180 latitude.
            wind_xarray["latitude"] = wind_xarray["latitude"] - 90

        wind_xarray["longitude"] = xarray.where(
            wind_xarray.ambiguity_select == 0, numpy.nan, wind_xarray.longitude
        )
        dsname = "WINDSPEED"

    if new_file:
        dsname = "ASCATUHR"
        wind_xarray["sigma0_mean"] = wind_xarray["sig"]

    # Old files were separate files for sigma (avewr) and windspeeds (WRave3)
    elif not new_file and "sig_fore" in wind_xarray.variables:
        dsname = "SIGMA"
        # bad vals where sigma 0 is -99.0
        wind_xarray["latitude"] = (
            xarray.where(wind_xarray.sig_fore < -98.0, numpy.nan, wind_xarray.latitude)
            - 90
        )

        wind_xarray["latitude"] = xarray.where(
            wind_xarray.latitude < -90.0, numpy.nan, wind_xarray.latitude
        )

        wind_xarray["longitude"] = xarray.where(
            wind_xarray.sig_fore < -98.0, numpy.nan, wind_xarray.longitude
        )
        wind_xarray["sigma0_mean"] = (
            wind_xarray["sig_fore"] + wind_xarray["sig_aft"] + wind_xarray["sig_mid"]
        ) / 3

    if "time" in wind_xarray.variables:
        # These files are not correct yet.  Pull YYYYMMDD from filename for now,
        # set hour to 1200.
        # Just to get something to plot.
        expected_yyyymmdd = os.path.basename(wind_xarray.source_file_names[0]).split(
            "_"
        )[1]
        dt = datetime.strptime(expected_yyyymmdd + "0000", "%Y%m%d%H%M")
        # Find the "true" minimum valid time of the data array.
        # Simply running min() on the time data array will
        # return "2000-01-01T00:00:00". The second lowest
        # value represents the true start time of the data
        # numpy.unique is a quick and easy way to do this,
        # it will return a sorted array of unique times.
        native_min_time = numpy.unique(wind_xarray["time"])[1]
        if abs((native_min_time - numpy.datetime64(dt)) * 1e-9) > 86400:
            # If the the native time difference is less than a day compared to the
            # expected, then assume the time array actually contains the correct dates!
            # Otherwise, time offset adjustments are needed
            timediff = numpy.datetime64(dt) - numpy.datetime64("2000-01-01T00:00:00")
            min_time = native_min_time + timediff
            min_time_diff = min_time - numpy.datetime64(dt)
            if min_time_diff * 1e-9 > 86400:
                LOG.info(
                    "Start time greater than 1day ahead of expected time. "
                    "Removing one day from applied offset"
                )
                timediff -= numpy.array(timedelta(days=1), dtype="timedelta64")
            wind_xarray["time"] = wind_xarray["time"] + timediff

        # This is all it will be in the end
        # wind_xarray = wind_xarray.rename({'time', 'time'})
        wind_xarray["time"] = xarray.where(
            wind_xarray["wind_speed_kts"] >= 0,
            wind_xarray["time"],
            numpy.datetime64("NaT"),
        )

    else:
        startdt = datetime.strptime(wind_xarray.SZF_start_time[:-1], "%Y%m%d%H%M%S")
        enddt = datetime.strptime(wind_xarray.SZF_stop_time[:-1], "%Y%m%d%H%M%S")
        middt = startdt + (enddt - startdt) / 2
        timearray = numpy.ma.array(
            numpy.zeros(shape=wind_xarray.latitude.shape).astype(int)
            + numpy.datetime64(middt)
        )
        wind_xarray["time"] = xarray.DataArray(
            timearray, name="time", coords=wind_xarray["latitude"].coords
        )
        wind_xarray = wind_xarray.set_coords(["time"])

    return {dsname: wind_xarray}


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read ASCAT UHR derived winds or normalized radar cross section data.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # Only SAR reads multiple files
    fname = fnames[0]
    wind_xarray = xarray.open_dataset(str(fname))
    wind_xarray.attrs["source_name"] = "unknown"
    wind_xarray.attrs["platform_name"] = "unknown"

    wind_xarray.attrs["source_file_names"] = [os.path.basename(fname)]
    wind_xarray.attrs["interpolation_radius_of_influence"] = 20000
    # 1.25km grid, 4km accuracy
    wind_xarray.attrs["sample_distance_km"] = 4
    wind_xarray.attrs["data_provider"] = "byu"
    wind_xarray.attrs["minimum_coverage"] = 20

    LOG.info("Read data from %s", fname)

    if (
        hasattr(wind_xarray, "institution")
        and "Brigham Young University" in wind_xarray.institution
    ):
        wind_xarrays = read_byu_data(wind_xarray, fname)

    for wind_xarray in wind_xarrays.values():
        LOG.info("Setting standard metadata")
        wind_xarray.attrs["start_datetime"] = get_min_from_xarray_time(
            wind_xarray, "time"
        )
        wind_xarray.attrs["end_datetime"] = get_max_from_xarray_time(
            wind_xarray, "time"
        )

        if "wind_speed_kts" in wind_xarray.variables:
            # These text files store wind speeds natively in kts
            wind_xarray["wind_speed_kts"].attrs["units"] = "kts"

        LOG.info(
            "Read data %s start_dt %s source %s platform %s data_provider %s roi"
            " %s native resolution",
            wind_xarray.attrs["start_datetime"],
            wind_xarray.attrs["source_name"],
            wind_xarray.attrs["platform_name"],
            wind_xarray.attrs["data_provider"],
            wind_xarray.attrs["interpolation_radius_of_influence"],
            wind_xarray.attrs["sample_distance_km"],
        )

    wind_xarrays["METADATA"] = wind_xarray[[]]
    if (
        wind_xarrays["METADATA"].start_datetime == wind_xarrays["METADATA"].end_datetime
        and "SZF_start_time" in wind_xarray.attrs
    ):
        # Use alternate attributes to set start and end datetime
        from datetime import datetime

        wind_xarrays["METADATA"].attrs["start_datetime"] = datetime.strptime(
            wind_xarray.SZF_start_time, "%Y%m%d%H%M%SZ"
        )
        wind_xarrays["METADATA"].attrs["end_datetime"] = datetime.strptime(
            wind_xarray.SZF_stop_time, "%Y%m%d%H%M%SZ"
        )

    return wind_xarrays



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read pre-processed GeoIPS-formatted NetCDF data."""

# Python Standard Libraries
from datetime import datetime
from os.path import basename
import logging

# Third-Party Libraries
import xarray

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "geoips_netcdf"
source_names = ["any"]


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read preprocessed geoips netcdf output.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    LOG.info("Reading files %s", fnames)

    xarray_objs = {}
    for fname in fnames:
        xarray_objs[basename(fname)] = read_xarray_netcdf(fname)

    xarray_objs["METADATA"] = list(xarray_objs.values())[0][[]]

    return xarray_objs


def read_xarray_netcdf(ncdf_fname):
    """Read NetCDF file written out using the xarray python package."""
    try:
        xarray_obj = xarray.open_dataset(ncdf_fname)
    except IOError:
        raise IOError
    for attr in xarray_obj.attrs.keys():
        if "datetime" in attr:
            xarray_obj.attrs[attr] = datetime.strptime(xarray_obj.attrs[attr], "%c")
        if attr == "None":
            xarray_obj.attrs[attr] = None
        if attr == "True":
            xarray_obj.attrs[attr] = True
        if attr == "False":
            xarray_obj.attrs[attr] = False
    return xarray_obj



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Procflows interface module."""

from geoips.interfaces.base import BaseModuleInterface


class ProcflowsInterface(BaseModuleInterface):
    """Class-based interface for processing workflows (procflows).

    Proclows drive a specific collection of steps for a particular type of
    processing.  Currently available are:

    * single_source: single input type and single output type.
    * overlay: two input types (one for foreground and one for background),
      with a single output type.
    * config_based: efficient method for producing all possible outputs for a
      given set of data files.
    """

    name = "procflows"
    required_args = {"standard": ["fnames"]}
    required_kwargs = {"standard": ["command_line_args"]}


procflows = ProcflowsInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard GeoIPS xarray dictionary based ABI NetCDF data reader."""

# Python Standard Libraries
import logging
import os
from glob import glob
from datetime import datetime, timedelta

# Third-Party Libraries
import numpy as np
from scipy.ndimage import zoom
import xarray

from geoips.interfaces import readers
from geoips.utils.context_managers import import_optional_dependencies
from geoips.plugins.modules.readers.utils.geostationary_geolocation import (
    get_geolocation_cache_filename,
    get_geolocation,
    AutoGenError,
)


LOG = logging.getLogger(__name__)
# np.seterr(all='raise')

# Installed Libraries

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package & print to LOG.info if the import fails."""
    # If this reader is not installed on the system, don't fail alltogether,
    # just skip this import. This reader will not work if the import fails
    # and the package will have to be installed to process data of this type.
    import netCDF4 as ncdf
    import numexpr as ne

interface = "readers"
family = "standard"
name = "abi_netcdf"
source_names = ["abi"]

nprocs = 6

try:
    ne.set_num_threads(nprocs)
except Exception:
    LOG.info(
        f"Failed numexpr.set_num_threads in {__file__}. "
        "If numexpr is not installed and you need it, install it."
    )

DONT_AUTOGEN_GEOLOCATION = False
if os.getenv("DONT_AUTOGEN_GEOLOCATION"):
    DONT_AUTOGEN_GEOLOCATION = True

# These should be added to the data file object
BADVALS = {
    "Off_Of_Disk": -999.9,
    "Conditional": -999.8,
    "Out_Of_Valid_Range": -999.7,
    "No_Value": -999.6,
    "Unitialized": -9999.9,
}

DATASET_INFO = {
    "MED": ["B01", "B03", "B05"],
    "HIGH": ["B02"],
    "LOW": [
        "B04",
        "B06",
        "B07",
        "B08",
        "B09",
        "B10",
        "B11",
        "B12",
        "B13",
        "B14",
        "B15",
        "B16",
    ],
}
ALL_GEO_VARS = [
    "solar_zenith_angle",
    "satellite_zenith_angle",
    "solar_azimuth_angle",
    "satellite_azimuth_angle",
    "latitude",
    "longitude",
]
ALL_CHANS = {
    "LOW": [
        "B04Rad",
        "B04Ref",  # 1.37um Near-IR Cirrus
        "B06Rad",
        "B06Ref",  # 2.2um  Near-IR Cloud Particle Size
        "B07Rad",
        "B07BT",  # 3.9um  IR      Shortwave Window
        "B08Rad",
        "B08BT",  # 6.2um  IR      Upper-level tropospheric water vapor
        "B09Rad",
        "B09BT",  # 6.9um  IR      Mid-level water vapor
        "B10Rad",
        "B10BT",  # 7.3um  IR      Lower-level Water Vapor
        "B11Rad",
        "B11BT",  # 8.4um  IR      Cloud-top phase
        "B12Rad",
        "B12BT",  # 9.6um  IR      Ozone
        "B13Rad",
        "B13BT",  # 10.3um IR      Clean IR Longwave Window
        "B14Rad",
        "B14BT",  # 11.2um IR      IR Longwave window
        "B15Rad",
        "B15BT",  # 12.3um IR      Dirty Longwave Window
        "B16Rad",
        "B16BT",
    ],  # 13.3um IR      CO2 Longwave infrared
    "MED": [
        "B01Rad",
        "B01Ref",  # 0.47um Vis     Blue
        "B02Rad",
        "B02Ref",  # 0.64um Vis     Red
        "B05Rad",
        "B05Ref",
    ],  # 1.6um  Near-IR Snow/Ice
    "HIGH": ["B03Rad", "B03Ref"],  # 0.86um Near-IR Veggie
}


def metadata_to_datetime(metadata):
    """Use information from the metadata to get the image datetime."""
    times = metadata["var_info"]["time_bounds"]
    epoch = datetime(2000, 1, 1, 12, 0, 0)
    start_time = epoch + timedelta(seconds=times[0])
    end_time = epoch + timedelta(seconds=times[1])
    return start_time, end_time


def _get_files(path):
    """Get a list of file names from the input path."""
    if os.path.isfile(path):
        fnames = [path]
    elif os.path.isdir(path):
        fnames = glob(os.path.join(path, "OR_ABI*.nc"))
    elif os.path.isfile(path[0]):
        fnames = path
    else:
        raise IOError("No such file or directory: {0}".format(path))
    return fnames


def _check_file_consistency(metadata):
    """
    Check that all input metadata are from the same image time.

    Performs cheks on platform_ID, instrument_type, processing_level, and times.
    If these are all equal, returns True.
    If any differ, returns False.
    """
    # Checking file-level metadata for exact equality in the following fields
    # Was failing on processing_level due to extra NASA text sometimes
    # Also fails on time_coverage_end due to deviations on the order of 0.5 seconds
    #   Probably should fashion better checks for these two.
    # checks = ['platform_ID', 'instrument_type', 'processing_level', 'timeline_id',
    #           'scene_id', 'time_coverage_start', 'time_coverage_end']
    checks = [
        "platform_ID",
        "instrument_type",
        "timeline_id",
        "scene_id",
        "time_coverage_start",
    ]
    for name in checks:
        check_set = set(
            [metadata[fname]["file_info"][name] for fname in metadata.keys()]
        )
        if len(check_set) != 1:
            LOG.debug("Failed on {0}. Found: {1}".format(name, check_set))
            return False
    return True


def _get_file_metadata(df):
    """Gather all of the file-level metadata."""
    metadata = {}
    md_names = [
        "id",
        "dataset_name",
        "naming_authority",
        "institution",
        "project",
        "iso_series_metadata_id",
        "Conventions",
        "Metadata_Conventions",
        "keywords_vocabulary",
        "standard_name_vocabulary",
        "title",
        "summary",
        "license",
        "keywords",
        "cdm_data_type",
        "orbital_slot",
        "platform_ID",
        "instrument_type",
        "processing_level",
        "date_created",
        "production_site",
        "production_environment",
        "production_data_source",
        "timeline_id",
        "scene_id",
        "spatial_resolution",
        "time_coverage_start",
        "time_coverage_end",
    ]
    metadata_dict = {"id": "instrument_ID"}
    for name in md_names:
        try:
            if hasattr(df, name):
                metadata[name] = getattr(df, name)
            else:
                metadata[name] = getattr(df, metadata_dict[name])
        except AttributeError:
            LOG.info("Warning! File-level metadata field missing: {0}".format(name))
    return metadata


def _get_variable_metadata(df):
    """
    Gather all required variable level metadata.

    Some are skipped or gathered later as needed.
    """
    metadata = {}
    # Note: We have skipped DQF, Rad, band_wavelength_star_look, num_star_looks,
    #       star_id, t, t_star_look, x_image, x_image_bounds, y_image, y_image_bounds,
    #       geospatial_lat_lon_extent, goes_imager_projection
    md_names = [
        "band_id",
        "band_wavelength",
        "earth_sun_distance_anomaly_in_AU",
        "esun",
        "kappa0",
        "max_radiance_value_of_valid_pixels",
        "min_radiance_value_of_valid_pixels",
        "missing_pixel_count",
        "nominal_satellite_height",
        "nominal_satellite_subpoint_lat",
        "nominal_satellite_subpoint_lon",
        "percent_uncorrectable_L0_errors",
        "planck_bc1",
        "planck_bc2",
        "planck_fk1",
        "planck_fk2",
        "processing_parm_version_container",
        "saturated_pixel_count",
        "std_dev_radiance_value_of_valid_pixels",
        "time_bounds",
        "undersaturated_pixel_count",
        "valid_pixel_count",
        "x",
        "y",
        "yaw_flip_flag",
    ]
    for name in md_names:
        try:
            metadata[name] = df.variables[name][...]
            if metadata[name].size == 1:
                metadata[name] = metadata[name][()]
            if name in ["x", "y"]:
                metadata[f"{name}_add_offset"] = df.variables[name].add_offset
                metadata[f"{name}_scale_factor"] = df.variables[name].scale_factor
        except KeyError:
            LOG.info("Warning! Variable-level metadata field missing: {0}".format(name))
    metadata["num_lines"] = metadata["y"].size
    metadata["num_samples"] = metadata["x"].size
    return metadata


def _get_lat_lon_extent_metadata(df):
    """Get lat lon extent metadata."""
    glle = df.variables["geospatial_lat_lon_extent"]
    metadata = {}
    md_names = [
        "geospatial_eastbound_longitude",
        "geospatial_lat_center",
        "geospatial_lat_nadir",
        "geospatial_lon_center",
        "geospatial_lon_nadir",
        "geospatial_northbound_latitude",
        "geospatial_southbound_latitude",
        "geospatial_westbound_longitude",
    ]
    for name in md_names:
        try:
            metadata[name] = getattr(glle, name)
            if metadata[name].size == 1:
                metadata[name] = metadata[name][()]
        except AttributeError:
            LOG.info("Warning! Lat lon extent metadata field missing: {0}".format(name))
    return metadata


def _get_imager_projection(df):
    """Get imager projection."""
    gip = df.variables["goes_imager_projection"]
    metadata = {}
    md_names = [
        "inverse_flattening",
        "latitude_of_projection_origin",
        "longitude_of_projection_origin",
        "perspective_point_height",
        "semi_major_axis",
        "semi_minor_axis",
    ]
    for name in md_names:
        try:
            metadata[name] = getattr(gip, name)
            if metadata[name].size == 1:
                metadata[name] = metadata[name][()]
        except AttributeError:
            LOG.info("Warning! Lat lon extent metadata field missing: {0}".format(name))
    metadata["grid_mapping"] = getattr(gip, "grid_mapping_name")
    return metadata


def _get_metadata(df, fname, **kwargs):
    """
    Gather metadata for the data file and return as a dictionary.

    Note: We are gathering all of the available metadata in case it is needed at
    some point.
    """
    metadata = {}
    # Gather all file-level metadata
    metadata["file_info"] = _get_file_metadata(df)
    # Gather all variable-level metadata
    metadata["var_info"] = _get_variable_metadata(df)
    # Gather lat lon extent info
    metadata["ll_info"] = _get_lat_lon_extent_metadata(df)
    # Gather projection info
    metadata["projection"] = _get_imager_projection(df)
    # Gather some useful info to the top level
    try:
        metadata["path"] = df.filepath()
    except ValueError:
        # Without cython installed, df.filepath() does not work
        metadata["path"] = fname
    metadata["satellite"] = metadata["file_info"]["platform_ID"]
    metadata["sensor"] = "ABI"
    metadata["num_lines"] = metadata["var_info"]["y"].size
    metadata["num_samples"] = metadata["var_info"]["x"].size
    return metadata


def get_latitude_longitude(metadata, BADVALS, sect=None):
    """
    Get latitudes and longitudes.

    This routine accepts a dictionary containing metadata as read from a NCDF4
    format file, and returns latitudes and longitudes for a full disk.
    """
    # If the filename format needs to change for the pre-generated geolocation
    # files, please discuss prior to changing.  It will force recreation of all
    # files, which can be problematic for large numbers of sectors
    fname = get_geolocation_cache_filename("GEOLL", metadata)
    if not os.path.isfile(fname):
        if sect is not None and DONT_AUTOGEN_GEOLOCATION and "tc2019" not in sect.name:
            msg = (
                f"GETGEO Requested NO AUTOGEN GEOLOCATION. "
                f"Could not create latlonfile for ad {metadata['scene']}: {fname}"
            )
            LOG.error(msg)
            raise AutoGenError(msg)

        LOG.debug("Calculating latitudes and longitudes.")

        r2d = 180.0 / np.pi  # NOQA

        lambda0 = np.radians(metadata["lon0"])  # NOQA
        Re = metadata["Re"]
        # invf = metadata['invf']
        Rp = metadata["Rp"]
        # e = np.sqrt((1 / invf) * (2 - 1 / invf))
        H = metadata["H_m"]
        c = H**2 - Re**2  # NOQA

        # Python 3 netcdf reads create a masked array, while Python 2 netcdf reads
        # create ndarray. These should NOT be masked, so if we have a masked array,
        # fill it.
        if isinstance(metadata["x"], np.ma.core.MaskedArray):
            x = np.float64(metadata["x"].filled())
        else:
            x = np.float64(metadata["x"])
        if isinstance(metadata["y"], np.ma.core.MaskedArray):
            y = np.float64(metadata["y"].filled())
        else:
            y = np.float64(metadata["y"])

        LOG.info("      Making {0} by {1} grid.".format(x.size, y.size))
        # Need to transpose the latline, then repeat lonsize times
        yT = y[np.newaxis].T
        y = np.hstack([yT for num in range(x.size)])
        # Repeat lonline latsize times
        x = np.vstack([x for num in range(yT.size)])

        # Note: In this next section, we will be reusing memory space as much as
        #       possible. To make this as transparent as possible, we will do all
        #       variable assignment first, then fill them
        # This method requires that all lines remain in the SAME ORDER or things
        # will go very badly
        cosx = np.empty_like(x)
        cosy = np.empty_like(x)
        a = np.empty_like(x)
        b = np.empty_like(x)
        sinx = x  # X is not needed after the line that defines sinx
        siny = y  # Y is not needed after the line that defines siny
        rs = a
        sx = b
        sy = cosy  # sinx is not needed after the line that defines sy
        sz = cosx  # cosx is not needed after the line that defines sz
        lats = rs
        lons = sz

        LOG.info("      Calculating intermediate steps")
        Rrat = Re**2 / Rp**2  # NOQA
        ne.evaluate("cos(x)", out=cosx)  # NOQA
        ne.evaluate("cos(y)", out=cosy)  # NOQA
        ne.evaluate("sin(x)", out=sinx)  # NOQA
        ne.evaluate("sin(y)", out=siny)  # NOQA
        ne.evaluate("sinx**2 + cosx**2 * (cosy**2 + siny**2 * Rrat)", out=a)  # NOQA
        ne.evaluate("-2 * H * cosx * cosy", out=b)  # NOQA
        ne.evaluate("(-b - sqrt(b**2 - (4 * a * c))) / (2 * a)", out=rs)  # NOQA
        good_mask = np.isfinite(rs)

        ne.evaluate("rs * cosx * cosy", out=sx)  # NOQA
        ne.evaluate("rs * cosx * siny", out=sz)  # NOQA
        ne.evaluate("rs * sinx", out=sy)  # NOQA

        LOG.info("Calculating Latitudes")
        ne.evaluate("r2d * arctan(Rrat * sz / sqrt((H - sx)**2 + sy**2))", out=lats)
        LOG.info("Calculating Longitudes")
        lons = ne.evaluate("r2d * (lambda0 + arctan(sy / (H - sx)))", out=lons)
        lats[~good_mask] = BADVALS["Off_Of_Disk"]
        lons[~good_mask] = BADVALS["Off_Of_Disk"]
        LOG.info("Done calculating latitudes and longitudes")

        with open(fname, "w") as df:
            lats.tofile(df)
            lons.tofile(df)
        # # Possible switch to xarray based geolocation files, but we lose memmapping.
        # ds = xarray.Dataset(
        #     {
        #         'latitude':(['x','y'],lats),
        #         'longitude':(['x','y'],lons)
        #         }
        #     )
        # ds.to_netcdf(fname)

    # Create memmap to the lat/lon file
    # Nothing will be read until explicitly requested
    # We are mapping this here so that the lats and lons are available when
    # calculating satlelite angles
    LOG.info(
        "GETGEO memmap to {} : lat/lon file for {}".format(fname, metadata["scene"])
    )

    shape = (metadata["num_lines"], metadata["num_samples"])
    offset = 8 * metadata["num_samples"] * metadata["num_lines"]
    lats = np.memmap(fname, mode="r", dtype=np.float64, offset=0, shape=shape)
    lons = np.memmap(fname, mode="r", dtype=np.float64, offset=offset, shape=shape)
    # Possible switch to xarray based geolocation files, but we lose memmapping
    # saved_xarray = xarray.load_dataset(fname)
    # lons = saved_xarray['longitude'].to_masked_array()
    # lats = saved_xarray['latitude'].to_masked_array()

    return lats, lons


def _get_geolocation_metadata(metadata):
    """
    Gather all of the metadata used in creating geolocation data for input file.

    This is split out so we can easily create a chash of the data for creation
    of a unique filename. This allows us to avoid recalculation of angles that
    have already been calculated.
    """
    geomet = {}
    # G16 -> goes-16
    geomet["platform_name"] = metadata["file_info"]["platform_ID"].replace("G", "goes-")
    geomet["Re"] = metadata["projection"]["semi_major_axis"]
    geomet["Rp"] = metadata["projection"]["semi_minor_axis"]
    geomet["invf"] = metadata["projection"]["inverse_flattening"]
    geomet["e"] = 0.0818191910435
    geomet["pphgt"] = metadata["projection"]["perspective_point_height"]
    geomet["H_m"] = geomet["Re"] + geomet["pphgt"]
    geomet["lat0"] = metadata["projection"]["latitude_of_projection_origin"]
    geomet["lon0"] = metadata["projection"]["longitude_of_projection_origin"]
    geomet["scene"] = metadata["file_info"]["scene_id"]
    # Just getting the nadir resolution in kilometers.  Must extract from a string.
    geomet["res_km"] = float(
        metadata["file_info"]["spatial_resolution"].split()[0][0:-2]
    )
    geomet["roi_factor"] = 5  # roi = res * roi_factor, was 10
    geomet["num_lines"] = metadata["var_info"]["num_lines"]
    geomet["num_samples"] = metadata["var_info"]["num_samples"]
    geomet["x"] = metadata["var_info"]["x"]
    geomet["y"] = metadata["var_info"]["y"]
    return geomet


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """
    Read ABI NetCDF data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    return readers.read_data_to_xarray_dict(
        fnames,
        call_single_time,
        metadata_only,
        chans,
        area_def,
        self_register,
    )


def call_single_time(
    fnames, metadata_only=False, chans=None, area_def=None, self_register=False
):
    """
    Read ABI NetCDF data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    gvars = {}
    datavars = {}
    standard_metadata = {}
    adname = "undefined"
    if area_def and self_register:
        raise ValueError("area_def and self_register are mutually exclusive keywords")
    elif area_def:
        adname = area_def.area_id

    # Test inputs
    if area_def and self_register:
        raise ValueError(
            "sector_definition and self_register are mutually exclusive keywords"
        )

    # Get metadata for all input data files
    # Check to be sure that all input files are form the same image time
    all_metadata = {}
    for fname in fnames:
        if chans:
            gotone = False
            for chan in chans:
                currchan = (
                    chan.replace("BT", "")
                    .replace("Ref", "")
                    .replace("Rad", "")
                    .replace("B", "C")
                )
                if currchan in fname:
                    gotone = True
            if not gotone:
                LOG.info(
                    "SKIPPING file %s, not needed from channel list %s", fname, chans
                )
                continue
        try:
            with ncdf.Dataset(str(fname), "r") as df:
                all_metadata[fname] = _get_metadata(df, fname)
        except IOError as resp:
            LOG.exception("BAD FILE %s skipping", resp)
            continue
        if metadata_only:
            LOG.info("Only need metadata from first file, returning")
            break

    if not _check_file_consistency(all_metadata):
        raise ValueError("Input files inconsistent.")

    # Now put together a dict that shows what we actually got
    # This is largely to help us read in order and only create arrays of the minimum
    # required size.
    # Dict structure is channel{metadata}
    file_info = {}

    xarray_obj = xarray.Dataset()
    for md in all_metadata.values():
        ch = "B{0:02d}".format(int(md["var_info"]["band_id"]))
        if ch not in file_info:
            file_info[ch] = md

    xarray_obj.attrs["file_metadata"] = file_info.copy()

    # Most of the metadata are the same between files.
    # From here on we will just rely on the metadata from a single data file
    # for each resolution
    res_md = {}
    for res in ["LOW", "MED", "HIGH"]:
        # Find a file for this resolution: Any one will do
        res_chans = list(set(DATASET_INFO[res]).intersection(file_info.keys()))
        if res_chans:
            res_md[res] = file_info[res_chans[0]]

    # If we plan to self register, make sure we requested a resolution that we actually
    # plan to read.
    #
    # This could be problematic if we try to self-register to LOW when only
    # reading MED or something
    if self_register and self_register not in res_md:
        raise ValueError(
            "Resolution requested for self registration has not been read."
        )

    # Gather metadata
    # Assume the same for all resolutions.  Not true, but close enough.
    highest_md = res_md[list(res_md.keys())[0]]
    sdt, edt = metadata_to_datetime(highest_md)
    xarray_obj.attrs["start_datetime"] = sdt
    xarray_obj.attrs["end_datetime"] = edt
    xarray_obj.attrs["source_name"] = "abi"
    xarray_obj.attrs["data_provider"] = "noaa"
    xarray_obj.attrs["source_file_names"] = [
        os.path.basename(fname) for fname in fnames
    ]

    # G16 -> goes-16
    xarray_obj.attrs["platform_name"] = highest_md["file_info"]["platform_ID"].replace(
        "G", "goes-"
    )
    xarray_obj.attrs["area_definition"] = area_def

    # Get appropriate sector name
    if area_def:
        xarray_obj.attrs["area_id"] = area_def.area_id
        xarray_obj.attrs["area_definition"] = area_def
    else:
        if self_register and self_register not in DATASET_INFO:
            raise ValueError(
                f"Unrecognized resolution name requested for self "
                f"registration: {self_register}"
            )
        scene_id = highest_md["file_info"]["dataset_name"].split("_")[1].split("-")[2]
        if scene_id == "RadF":
            xarray_obj.attrs["area_id"] = "Full-Disk"
        elif scene_id == "RadM1":
            xarray_obj.attrs["area_id"] = "Mesoscale-1"
        elif scene_id == "RadM2":
            xarray_obj.attrs["area_id"] = "Mesoscale-2"
        elif scene_id == "RadC":
            xarray_obj.attrs["area_id"] = "CONUS"
        xarray_obj.attrs["area_definition"] = None
    adname = xarray_obj.attrs["area_id"]

    if metadata_only:
        LOG.info("Only need metadata from first file, returning")
        return {"METADATA": xarray_obj}

    # Create list of all possible channels for the case where no channels were requested
    all_chans_list = []
    for chl in ALL_CHANS.values():
        all_chans_list += chl

    # If specific channels were requested, check them against the input data
    # If specific channels were requested, but no files exist for one of the
    # channels, then error
    if chans:
        for chan in chans:
            if chan in ALL_GEO_VARS:
                continue
            if chan not in all_chans_list:
                raise ValueError("Requested channel {0} not recognized.".format(chan))
            if chan[0:3] not in file_info.keys():
                continue
                # raise ValueError(
                #     'Requested channel {0} not found in input data.'.format(chan)
                # )

    # If no specific channels were requested, get everything
    if not chans:
        chans = all_chans_list

    # Creates dict whose keys are band numbers in the form B## and whose values are
    # lists containing the data types(s) requested for the band (e.g. Rad, Ref, BT).
    chan_info = {}
    for ch in chans:
        if ch in ALL_GEO_VARS:
            continue
        chn = ch[0:3]
        typ = ch[3:]
        if chn not in chan_info:
            chan_info[chn] = []
        chan_info[chn].append(typ)

    # Gather geolocation data
    # Assume datetime the same for all resolutions.  Not true, but close enough.
    # This save us from having very slightly different solar angles for each channel.
    # Loop over resolutions and get metadata as needed
    if self_register:
        LOG.info("")
        LOG.info(
            "Getting geolocation information for resolution {} for {}.".format(
                self_register, adname
            )
        )
        # Get just the metadata we need
        standard_metadata[adname] = _get_geolocation_metadata(res_md[self_register])
        fldk_lats, fldk_lons = get_latitude_longitude(
            standard_metadata[adname], BADVALS, area_def
        )
        gvars[adname] = get_geolocation(
            sdt, standard_metadata[adname], fldk_lats, fldk_lons, BADVALS, area_def
        )
        if not gvars[adname]:
            LOG.error(
                f"GEOLOCATION FAILED for adname {adname} DONT_AUTOGEN_GEOLOCATION "
                f"is: {DONT_AUTOGEN_GEOLOCATION}"
            )
            gvars[adname] = {}
    else:
        for res in ["LOW", "MED", "HIGH"]:
            try:
                res_md[res]
            except KeyError:
                continue
            try:
                # Get just the metadata we need
                standard_metadata[res] = _get_geolocation_metadata(res_md[res])
                fldk_lats, fldk_lons = get_latitude_longitude(
                    standard_metadata[res], BADVALS, area_def
                )
                gvars[res] = get_geolocation(
                    sdt, standard_metadata[res], fldk_lats, fldk_lons, BADVALS, area_def
                )
            except ValueError as resp:
                LOG.error(
                    "{} GEOLOCATION FAILED FOR {} resolution {}. Skipping.".format(
                        resp, adname, res
                    )
                )
            # Not sure what this does, but it fails if we only pass one resolution
            # worth of data...
            #
            # MLS I think this may be necessary to avoid catastrophic failure if
            # geolocation has not been autogenerated.  I'll see.
            # if not gvars[res]:
            #     LOG.error(
            #         f"GEOLOCATION FAILED for {adname} resolution {res} "
            #         f"DONT_AUTOGEN_GEOLOCATION is: {DONT_AUTOGEN_GEOLOCATION}"
            #     )
            #     gvars[res] = {}

    LOG.interactive("Done with geolocation for {}".format(adname))
    LOG.info("")

    # Read the data
    # Will read all data if sector_definition is None
    # Will only read required data if an sector_definition is provided
    for chan, types in chan_info.items():
        # If we didn't pass all the channel data, skip non-existent data types
        if chan not in file_info:
            continue
        LOG.info("Reading {}".format(chan))
        chan_md = file_info[chan]
        for res, res_chans in DATASET_INFO.items():
            if chan in res_chans:
                break
        if (not self_register) and (
            res not in gvars.keys() or not gvars[res] or "latitude" not in gvars[res]
        ):
            LOG.info(
                f"We don't have geolocation information for {res} for {adname} "
                f"skipping {chan}."
            )
            continue
        if not area_def:
            dsname = res
        else:
            dsname = adname

        rad = ref = bt = False
        if "Rad" in types:
            rad = True
        if "Ref" in types:
            ref = True
        if "BT" in types:
            bt = True
        if self_register:
            data = get_data(chan_md, gvars[adname], rad, ref, bt)
        else:
            data = get_data(chan_md, gvars[res], rad, ref, bt)
        for typ, val in data.items():
            if dsname not in datavars:
                datavars[dsname] = {}
            datavars[dsname][chan + typ] = val

    # This needs to be fixed:
    #   remove any unneeded datasets from datavars and gvars
    #   also mask any values below -999.0
    if area_def:
        for res in ["LOW", "MED", "HIGH"]:
            if adname not in gvars and res in gvars and gvars[res]:
                gvars[adname] = gvars[res]
            try:
                gvars.pop(res)
            except KeyError:
                pass

    # NOTE: Moved zoom and subsample to get_data, but still need to move data
    #       to correct dataset and drop extra datasets
    if self_register:
        # Determine which resolution has geolocation
        LOG.info("Registering to {}".format(self_register))
        if self_register == "HIGH":
            datavars[adname] = datavars.pop("HIGH")
            for varname, var in datavars["LOW"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = zoom(var, 4, order=0)
            datavars.pop("LOW")
            for varname, var in datavars["MED"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = zoom(var, 2, order=0)
            datavars.pop("MED")

        elif self_register == "MED":
            datavars[adname] = datavars.pop("MED")
            for varname, var in datavars["LOW"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = zoom(var, 2, order=0)
            datavars.pop("LOW")
            for varname, var in datavars["HIGH"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = var[::2, ::2]
            datavars.pop("HIGH")

        elif self_register == "LOW":
            datavars[adname] = datavars.pop("LOW")
            for varname, var in datavars["MED"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = var[::2, ::2]
            datavars.pop("MED")
            for varname, var in datavars["HIGH"].items():
                datavars[adname][varname] = var
                # datavars[adname][varname] = var[::4, ::4]
            datavars.pop("HIGH")

        else:
            raise ValueError("No geolocation data found.")

    # basically just reformat the all_metadata dictionary to
    # reference channel names as opposed to file names..
    # band_metadata = get_band_metadata(all_metadata)

    # Remove lines and samples arrays.  Not needed.
    for res in gvars.keys():
        try:
            LOG.info(f"Popping Lines and samples for {res}")
            gvars[res].pop("Lines")
            gvars[res].pop("Samples")
        except KeyError:
            LOG.info(f"Did not find Lines and Samples to pop for {res}")
            pass
        try:
            LOG.info(f"Masking greater than 75 degrees sat zenith angle for {res}")
            for varname, var in gvars[res].items():
                LOG.info(f"Masking {varname} greater than 75 degrees sat zenith angle")
                gvars[res][varname] = np.ma.array(
                    var, mask=gvars[res]["satellite_zenith_angle"].mask
                )
                gvars[res][varname] = np.ma.masked_where(
                    gvars[res]["satellite_zenith_angle"] > 75, gvars[res][varname]
                )
        except KeyError:
            LOG.info(f"Did not find satellite zenith angle to mask for {res}")
            pass
    for ds in datavars.keys():
        if not datavars[ds]:
            datavars.pop(ds)

    xarray_objs = {}
    for dsname in datavars.keys():
        xobj = xarray.Dataset()
        xobj.attrs = xarray_obj.attrs.copy()
        for varname in datavars[dsname].keys():
            xobj[varname] = xarray.DataArray(datavars[dsname][varname])
        for varname in gvars[dsname].keys():
            xobj[varname] = xarray.DataArray(gvars[dsname][varname])

        roi = 500
        roi_factor = 5
        if hasattr(xobj, "area_definition") and xobj.area_definition is not None:
            roi = max(
                xobj.area_definition.pixel_size_x, xobj.area_definition.pixel_size_y
            )
            LOG.info("Trying area_def roi %s", roi)
        for curr_res in standard_metadata.keys():
            if standard_metadata[curr_res]["res_km"] * 1000.0 * roi_factor > roi:
                roi = standard_metadata[curr_res]["res_km"] * 1000.0 * roi_factor
                LOG.info("Trying standard_metadata[%s] %s", curr_res, roi)
        xobj.attrs["interpolation_radius_of_influence"] = roi
        LOG.info(f"Using roi {roi}")
        xarray_objs[dsname] = xobj
        # At some point we may need to deconflict, but for now just use any of the
        # dataset attributes as the METADATA dataset
        xarray_objs["METADATA"] = xobj[[]]

    LOG.info("Done reading ABI data for %s", adname)
    LOG.info("")
    return xarray_objs


def get_band_metadata(all_metadata):
    """
    Get band metadata.

    This method basically just reformats the all_metadata
    dictionary that is set based on the metadata found
    in the netcdf object itself to reference channel
    names as opposed to filenames as the dictionary keys.
    """
    bandmetadata = {}
    for fname in all_metadata.keys():
        bandnum = all_metadata[fname]["var_info"]["band_id"][0]
        bandmetadata["B%02d" % bandnum] = {}
        bandmetadata["B%02d" % bandnum] = all_metadata[fname]
    return bandmetadata


def get_data(md, gvars, rad=False, ref=False, bt=False):
    """Read data for a full channel's worth of files."""
    # Coordinate arrays for reading
    if "Lines" in gvars and "Samples" in gvars:
        full_disk = False
        line_inds = gvars["Lines"]
        sample_inds = gvars["Samples"]
    else:
        full_disk = True

    band_num = md["var_info"]["band_id"]

    # Read radiance data for channel
    # For some reason, netcdf4 does not like it if you pass a bunch of indexes.
    # It is very slow and creates memory errors if there are too many.
    # Have to read ALL of the data, then subset.
    # Need to find a solution for this.
    # Use with to open the data file for reading and automatically close once complete
    with ncdf.Dataset(md["path"], "r") as df:
        if not full_disk:
            rad_data = np.float64(df.variables["Rad"][...][line_inds, sample_inds])
            qf = df.variables["DQF"][...][line_inds, sample_inds]
        else:
            # Here we need to determine which indexes to read based on the size of the
            # input geolocation data.  We assume that the geolocation data and the
            # variable data cover the same domain, just at a different resolution.
            geoloc_shape = np.array(gvars["solar_zenith_angle"].shape, dtype=np.float64)
            data_shape = np.array(df.variables["Rad"].shape, dtype=np.float64)
            # If the geolocation shape matches the data shape, just read the data
            if np.all(geoloc_shape == data_shape):
                rad_data = np.float64(df.variables["Rad"][...])
                qf = df.variables["DQF"][...]
            # Upscaling data to match geolocation shape
            elif np.all(np.maximum(geoloc_shape, data_shape) == geoloc_shape):
                # Data shape is smaller, so it is the denominator
                zoom_factor, remainder = np.divmod(geoloc_shape, data_shape)
                # Ensure that all elements of the zoom factor are whole numbers
                if np.any(remainder):
                    raise ValueError(
                        "Zoom factor is not a whole number.  "
                        "This section of code needs to be reexamined."
                    )
                # Ensure that all elements of the zoom factor are equal
                if not np.all(zoom_factor == zoom_factor[0]):
                    raise ValueError(
                        "Zoom factor must be equal for both dimensions.  "
                        "This section of code needs to be reexamined."
                    )
                # Perform the actual zoom
                zoom_factor = zoom_factor.astype(np.int)
                rad_data = zoom(
                    np.float64(df.variables["Rad"][...]), zoom_factor[0], order=0
                )
                qf = zoom(df.variables["DQF"][...], zoom_factor[0], order=0)
            # Downscaling data to match geolocation shape
            elif np.all(np.maximum(geoloc_shape, data_shape) == data_shape):
                # Geoloc shape is smaller, so it is the denominator
                zoom_factor, remainder = np.divmod(data_shape, geoloc_shape)
                # Ensure that all elements of the zoom factor are whole numbers
                if np.any(remainder):
                    raise ValueError(
                        "Zoom factor is not a whole number.  "
                        "This section of code needs to be reexamined."
                    )
                # Ensure that all elements of the zoom factor are equal
                if not np.all(zoom_factor == zoom_factor[0]):
                    raise ValueError(
                        "Zoom factor must be equal for both dimensions.  "
                        "This section of code needs to be reexamined."
                    )
                # Perform the actual subsampling
                LOG.info("Before zoom")
                zoom_factor = zoom_factor.astype(np.int)

                # NOTE: Strides are broken for netCDF4 library version < 4.6.2.
                #       At present, the most recent stable release is 4.6.1.
                #       Once 4.6.2 is released, this should be retested.
                #       See https://github.com/Unidata/netcdf4-python/issues/680

                rad_data = np.float64(
                    df.variables["Rad"][...][:: zoom_factor[0], :: zoom_factor[0]]
                )
                qf = df.variables["DQF"][...][:: zoom_factor[0], :: zoom_factor[0]]
                # rad_data = np.float64(
                #     df.variables['Rad'][::zoom_factor[0], ::zoom_factor[0]]
                # )
                # qf = df.variables['DQF'][::zoom_factor[0], ::zoom_factor[0]]
                LOG.info("After zoom")
            else:
                raise ValueError(
                    "Zoom factor cannot be computed.  "
                    "All either both dimensions of geolocation data must be "
                    "larger than the data dimensions or vice versa."
                )

    data = {"Rad": rad_data}
    # Originally was using -1, but this is a uint8 and set to 255 for off of disk
    data["Rad"][np.where(qf == 255)] = BADVALS["Off_Of_Disk"]
    data["Rad"][np.where(qf == 1)] = BADVALS["Conditional"]
    # This flag is ignored for now.  It masks good data that is useful for imagery.
    # I am unsure if this should be used in digital products.
    # data['Rad'][np.where(qf == 2)] = BADVALS['Out_Of_Valid_Range']
    data["Rad"][np.where(qf == 3)] = BADVALS["No_Value"]
    data["Rad"] = np.ma.masked_less_equal(data["Rad"], -999.0)

    # Get the bad data mask from radiances to reuse
    bad_data_mask = data["Rad"].mask

    # If reflectance is requested
    # Note, weird memory manipulations to save memory when radiances are not requested
    if ref:
        if band_num not in range(1, 7):
            raise ValueError(
                "Unable to calculate reflectances for band #{0}".format(band_num)
            )

        # Get the radiance data
        # Have to do this when using numexpr
        rad_data = data["Rad"][~bad_data_mask]  # NOQA

        # If we don't need radiances, then reuse the memory
        if not rad:
            data["Ref"] = data.pop("Rad")
        else:
            data["Ref"] = np.empty_like(data["Rad"])

        k0 = md["var_info"]["kappa0"]  # NOQA
        sun_zenith = gvars["solar_zenith_angle"][~bad_data_mask]  # NOQA
        # zoom_info = (
        #     np.array(rad_data.shape, dtype=np.float) /
        #     np.array(sun_zenith.shape, dtype=np.float
        # )
        # if zoom_info[0] == zoom_info[1] and int(zoom_info[0]) == zoom_info[0]:
        #     if zoom_info[0] > 1 and int(zoom_info[0]) == zoom_info[0]:
        #         sun_zenith = zoom(sun_zenith, zoom_info[0])
        #     elif zoom_info[0] < 1 and int(zoom_info[0]**-1) == zoom_info[0]**-1:
        #         sun_zenith = sun_zenith[::zoom_info[0]**-1, :: zoom_info[0]**-1]
        #     elif zoom_info[0] == 1:
        #         pass
        #     else:
        #         ValueError('Inappropriate zoom level calculated.')
        # else:
        #     ValueError('Inappropriate zoom level calculated.')

        deg2rad = np.pi / 180.0  # NOQA
        # NOTE:: ABI docs suggest doing solar zenith angle correction here, but
        #        since some algorithms (e.g. Fire-Temperature RGB) required uncorrected
        #        data, we will not do this here, instead leaving it to the algorithm
        #        developer to handle.
        data["Ref"][~bad_data_mask] = ne.evaluate("k0 * rad_data")

    if bt:
        if band_num not in range(7, 17):
            raise ValueError(
                "Unable to calculate brightness temperatures for band #{0}".format(
                    band_num
                )
            )

        # Get the radiance data
        # Have to do this when using numexpr
        rad_data = data["Rad"][~bad_data_mask]

        # Sometimes negative values occur due to oddities in calibration
        # These break the conversion to BT
        # Since these values are likely caused by very dark (cold) scenes
        #   that are below the sensor's sensitivity and we are not interested in
        #   having bad data scattered throughout the imager, we will set this
        #   to a very small value
        rad_data[rad_data <= 0] = 0.001

        # If we don't need radiances, then reuse the memory
        if not rad:
            data["BT"] = data.pop("Rad")
        else:
            data["BT"] = np.empty_like(data["Rad"])

        fk1 = md["var_info"]["planck_fk1"]  # NOQA
        fk2 = md["var_info"]["planck_fk2"]  # NOQA
        bc1 = md["var_info"]["planck_bc1"]  # NOQA
        bc2 = md["var_info"]["planck_bc2"]  # NOQA

        data["BT"][~bad_data_mask] = ne.evaluate(
            "(fk2 / log(fk1 / rad_data + 1) - bc1) / bc2"
        )

    # latitude is sometimes fully specified from area_def... So does not
    # relate to actual masked data..
    # This should only occur when area_def was provided and breaks when it wasn't.
    if "Lines" in gvars:
        usemask = np.ma.masked_less(gvars["Lines"], -990).mask
    else:
        usemask = gvars["latitude"].mask

    for varname in gvars.keys():
        if varname in ["latitude", "longitude"]:
            continue
        if hasattr(gvars[varname], "mask"):
            gvars[varname].mask = usemask
    for varname in data.keys():
        if hasattr(data[varname], "mask"):
            data[varname].mask = usemask
    return data
